
import numpy as np
from scipy.interpolate.interpolate import interp1d



def bin_fit(x,y, bins=10, kind='linear', bin_func=np.nanmean, bin_min_count=3, lower_upper='discard'):
    """Fit observations based on bin statistics
    
    Parameters
    ---------
    x : array_like
        x observations
    y : array_like
        y observations
    bins : int, array_like or (int, int)
        if int: <bins> binx evenly distributed on the x-axis
        if (xbins,ybins): <xbins> and <ybins> evenly distributed on the x and y axis respectively\n
        Note that ybins only make sense if every y-value maps to a single x-value
    kind : int or string
        degree of polynomial for fit (argument passed to scipy.interpolate.interpolate.interp1d)
    bin_func : function, optional
        Statistic function to apply on bins, default is nanmean
    bin_min_count : int, optional
        Minimum number of observations in bins to include
        Default is 3
    lower_upper : str, int, (str,str), (int,int)
        How to handle observations below and above first and last bin values. Can be:\n
        - "discard":
        - "extrapolate":
        - int: Set f(max(x)) to mean of first/last int observations
       
    
    Returns
    -------
    bin_x, fit_function
        
    """
    x, y = np.array(x[:]), np.array(y[:])
    if isinstance(bins, int):
        bins = np.linspace(np.nanmin(x), np.nanmax(x) + 1e-10, bins + 1)
    elif isinstance(bins, tuple) and len(bins)==2 and isinstance(bins[0], int) and isinstance(bins[1], int):
        xbins, ybins = bins
        if xbins>0:
            xbinsx = np.linspace(np.nanmin(x), np.nanmax(x) + 1e-10, xbins + 1)
        else:
            xbinsx = []
        if ybins>0:
            x1, f1 = bin_fit(y,x, kind=1, bins=ybins)
            xbinsy = f1(x1)
        else:
            xbinsy = []
        #x2, f2 = bin_fit(x,y, kind=1, bins=xbins)
        bins = sorted(np.r_[xbinsx, xbinsy ])

    digitized = np.digitize(x, bins)
    digitized[np.isnan(x) | np.isnan(y)] = -1

    masks = [digitized == i for i in range(1, len(bins))]
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bin_x = np.array([np.nanmean(x[mask]) for mask in masks])
        bin_y = np.array([bin_func(y[mask]) for mask in masks])
        bin_count = np.array([np.sum(mask) for mask in masks])
    #bin_x_fit, bin_y = [b[bin_count >= bin_min_count] for b in [bin_x, bin_y]]
    bin_x_fit = bin_x
    m = np.isnan(bin_x_fit)
    bin_x_fit[m] = ((bins[:-1]+bins[1:])/2)[m]
    bin_y_fit = bin_y.copy()
    bin_y_fit[bin_count<bin_min_count]= np.nan

    
    if isinstance(lower_upper, (str, int)):
        lower = upper = lower_upper
    else:
        lower, upper = lower_upper 
    
    #Add value to min(x)
    if bin_x_fit[0] > np.nanmin(x) or np.isnan(bin_y_fit[0]):
        if lower =='extrapolate':
        
            bin_y_fit = np.r_[bin_y_fit[0] - (bin_x_fit[0] - np.nanmin(x)) * (bin_y_fit[1] - bin_y_fit[0]) / (bin_x_fit[1] - bin_x_fit[0]), bin_y_fit]
            bin_x_fit = np.r_[np.nanmin(x), bin_x_fit]
        elif lower=="discard":
            pass
        elif isinstance(lower, int):
            bin_y_fit = np.r_[np.mean(y[~np.isnan(x)][np.argsort(x[~np.isnan(x)])[:lower]]), bin_y_fit]
            bin_x_fit = np.r_[np.nanmin(x), bin_x_fit]
        else:
            raise NotImplementedError("Argument for handling lower observations, %s, not implemented"%lower)
        
    #add value to max(x)
    if bin_x_fit[-1] < np.nanmax(x) or np.isnan(bin_y_fit[-1]):
        if upper == 'extrapolate':
            bin_y_fit = np.r_[bin_y_fit, bin_y_fit[-1] + (np.nanmax(x) - bin_x_fit[-1]) * (bin_y_fit[-1] - bin_y_fit[-2]) / (bin_x_fit[-1] - bin_x_fit[-2]) ]
            bin_x_fit = np.r_[bin_x_fit, np.nanmax(x)]
        elif upper=="discard":
            pass
        elif isinstance(upper, int):
            bin_y_fit = np.r_[bin_y_fit, np.mean(y[~np.isnan(x)][np.argsort(x[~np.isnan(x)])[-upper:]])]
            bin_x_fit = np.r_[bin_x_fit, np.nanmax(x)]
        else:
            raise NotImplementedError("Argument for handling upper observations, %s, not implemented"%upper)
        

    
    return bin_x_fit, _interpolate_fit(bin_x_fit, bin_y_fit, kind)
    
def perpendicular_bin_fit(x, y, bins = 30, fit_func=None, bin_min_count=3, plt=None):
    """Fit a curve to the values, (x,y) using bins that are perpendicular to an initial fit
    
    Parameters
    ---------
    x : array_like
        x observations
    y : array_like
        y observations
    bins : int
        Number of perpendicular bins
    fit_func : function(x,y) -> (x,y) or None
        Initial fit function
        If None, bin_fit with same number of bins are used
    bin_min_count : int, optional
        Minimum number of observations in bins to include
        Default is 3
    
    plt : pyplot or None
        If pyplot the fitting process is plotted on plt
        
    Returns
    -------
    fit_x, fit_y
    """
    
    if fit_func is None:
        fit_func = lambda x,y : bin_fit(x, y, bins, bin_func=np.nanmean)
    
    x,y = [v[~np.isnan(x)&~np.isnan(y)] for v in [x,y]]
    
    bfx,f = fit_func(x, y)
    bfy = f(bfx)
    
    
    bfx, bfy = [v[~np.isnan(bfx)&~np.isnan(bfy)] for v in [bfx,bfy]]
    if plt:
        x_range, y_range = [v.max()-v.min() for v in [x,y]]
        plt.ylim([y.min()-y_range*.1, y.max()+y_range*.1])
        plt.xlim([x.min()-x_range*.1, x.max()+x_range*.1])
        
    # divide curve into N segments of same normalized curve length 
    xg, xo = np.nanmax(bfx)-np.nanmin(bfx), np.nanmin(bfx)
    yg, yo = np.nanmax(bfy)-np.nanmin(bfy), np.nanmin(bfy)
    nbfx = (bfx-xo)/xg
    nbfy = (bfy-yo)/yg
    l = np.cumsum(np.sqrt(np.diff(nbfx)**2+np.diff(nbfy)**2))
    nx, ny = [np.interp(np.linspace(l[0], l[-1], bins+1), l, (xy[1:]+xy[:-1])/2) for xy in [nbfx,nbfy]]
    
    
    last = (-1,0)
    
    pc = []
    used = np.zeros_like(x).astype(np.bool)
    for i in range(0,len(nx)):
        i1,i2 = max(0,i-1), min(len(nx)-1,i+1)
        a =-(nx[i2]-nx[i1])/ (ny[i2]-ny[i1])
        b = (ny[i]-(a*nx[i]))*yg+yo
        a *=yg/xg
        x_ = [np.nanmin(x), np.nanmax(x)]
        m1 = np.sign(last[0])*y < np.sign(last[0])*((x-xo)*last[0]+last[1])
        m2 = np.sign(a)*y>np.sign(a)*(a*(x-xo)+b)
        m = m1&m2&~used
        if plt:
            plt.plot(x_, ((a)*(x_-xo))+b)
            plt.plot(x[m], y[m],'.')
        
        if np.sum(m)>=bin_min_count:
            pc.append((np.median(x[m]), np.median(y[m])))
            used = used|m
        last = (a,b)
    #bfx,bfy = zip(*pc)
    if plt:
        pbfx, pbfy = np.array(pc).T
        plt.plot(bfx,bfy, 'orange', label='initial_fit')
        plt.plot(pbfx, pbfy, 'gray', label="perpendicular fit")
        plt.legend()
    #PlotData(None, bfx,bfy)
    bin_x_fit, bin_y_fit = np.array(pc).T 
    return bin_x_fit, _interpolate_fit(bin_x_fit, bin_y_fit, kind="linear")

#Create mean function
def _interpolate_fit(bin_x_fit, bin_y_fit, kind='linear'):
    def fit(x):
        x = np.atleast_1d(x)[:].copy().astype(np.float)
        x[x<bin_x_fit[0]] = np.nan
        x[x>bin_x_fit[-1]] = np.nan
        m = ~(np.isnan(bin_x_fit)|np.isnan(bin_y_fit))
        return interp1d(bin_x_fit[m], bin_y_fit[m], kind)(x[:])
    return fit




import cython
import numpy as np
cimport numpy as np

import cython
import numpy as np
# cimport numpy as np


@cython.locals(p=cython.int, q=cython.int, f=cython.int, flow=list, k=cython.int, n=cython.int, ptr=cython.int)
def pair_range_amplitude(x):  # cpdef pair_range(np.ndarray[long,ndim=1]  x):
    """
    Returns a list of half-cycle-amplitudes
    x: Peak-Trough sequence (integer list of local minima and maxima)

    This routine is implemented according to
    "Recommended Practices for Wind Turbine Testing - 3. Fatigue Loads", 2. edition 1990, Appendix A
    except that a list of half-cycle-amplitudes are returned instead of a from_level-to_level-matrix
    """

    x = x - np.min(x)
    k = np.max(x)
    n = x.shape[0]
    S = np.zeros(n + 1)

    #A = np.zeros(k+1)
    flow = []
    S[1] = x[0]
    ptr = 1
    p = 1
    q = 1
    f = 0
    # phase 1
    while True:
        p += 1
        q += 1

        # read
        S[p] = x[ptr]
        ptr += 1

        if q == n:
            f = 1
        while p >= 4:
            if (S[p - 2] > S[p - 3] and S[p - 1] >= S[p - 3] and S[p] >= S[p - 2]) \
                or\
                    (S[p - 2] < S[p - 3] and S[p - 1] <= S[p - 3] and S[p] <= S[p - 2]):
                ampl = abs(S[p - 2] - S[p - 1])
                # A[ampl]+=2 #Two half cycles
                flow.append(ampl)
                flow.append(ampl)
                S[p - 2] = S[p]

                p -= 2
            else:
                break

        if f == 0:
            pass
        else:
            break
    # phase 2
    q = 0
    while True:
        q += 1
        if p == q:
            break
        else:
            ampl = abs(S[q + 1] - S[q])
            # A[ampl]+=1
            flow.append(ampl)
    return flow





@cython.locals(p=cython.int, q=cython.int, f=cython.int, flow=list, k=cython.int, n=cython.int, ptr=cython.int)
def pair_range_from_to(x):  # cpdef pair_range(np.ndarray[long,ndim=1]  x):
    """
    Returns a list of half-cycle-amplitudes
    x: Peak-Trough sequence (integer list of local minima and maxima)

    This routine is implemented according to
    "Recommended Practices for Wind Turbine Testing - 3. Fatigue Loads", 2. edition 1990, Appendix A
    except that a list of half-cycle-amplitudes are returned instead of a from_level-to_level-matrix
    """

    x = x - np.min(x)
    k = np.max(x)
    n = x.shape[0]
    S = np.zeros(n + 1)

    A = np.zeros((k + 1, k + 1))
    S[1] = x[0]
    ptr = 1
    p = 1
    q = 1
    f = 0
    # phase 1
    while True:
        p += 1
        q += 1

        # read
        S[p] = x[ptr]
        ptr += 1

        if q == n:
            f = 1
        while p >= 4:
            #print S[p - 3:p + 1]
            #print S[p - 2], ">", S[p - 3], ", ", S[p - 1], ">=", S[p - 3], ", ", S[p], ">=", S[p - 2], (S[p - 2] > S[p - 3] and S[p - 1] >= S[p - 3] and S[p] >= S[p - 2])
            #print S[p - 2], "<", S[p - 3], ", ", S[p - 1], "<=", S[p - 3], ", ", S[p], "<=", S[p - 2], (S[p - 2] < S[p - 3] and S[p - 1] <= S[p - 3] and S[p] <= S[p - 2])
            #print (S[p - 2] > S[p - 3] and S[p - 1] >= S[p - 3] and S[p] >= S[p - 2]) or (S[p - 2] < S[p - 3] and S[p - 1] <= S[p - 3] and S[p] <= S[p - 2])
            if (S[p - 2] > S[p - 3] and S[p - 1] >= S[p - 3] and S[p] >= S[p - 2]) or \
               (S[p - 2] < S[p - 3] and S[p - 1] <= S[p - 3] and S[p] <= S[p - 2]):
                A[S[p - 2], S[p - 1]] += 1
                A[S[p - 1], S[p - 2]] += 1
                S[p - 2] = S[p]
                p -= 2
            else:
                break

        if f == 1:
            break  # q==n
    # phase 2
    q = 0
    while True:
        q += 1
        if p == q:
            break
        else:
            #print S[q], "to", S[q + 1]
            A[S[q], S[q + 1]] += 1
    return A

@cython.locals(p=cython.int, q=cython.int, f=cython.int, flow=list, k=cython.int, n=cython.int, ptr=cython.int)
def pair_range_amplitude_mean(x):  # cpdef pair_range(np.ndarray[long,ndim=1]  x):
    """
    Returns a list of half-cycle-amplitudes
    x: Peak-Trough sequence (integer list of local minima and maxima)

    This routine is implemented according to
    "Recommended Practices for Wind Turbine Testing - 3. Fatigue Loads", 2. edition 1990, Appendix A
    except that a list of half-cycle-amplitudes are returned instead of a from_level-to_level-matrix
    """

    x = x - np.min(x)
    k = np.max(x)
    n = x.shape[0]
    S = np.zeros(n + 1)
    ampl_mean = []
    A = np.zeros((k + 1, k + 1))
    S[1] = x[0]
    ptr = 1
    p = 1
    q = 1
    f = 0
    # phase 1
    while True:
        p += 1
        q += 1

                # read
        S[p] = x[ptr]
        ptr += 1

        if q == n:
            f = 1
        while p >= 4:
            if (S[p - 2] > S[p - 3] and S[p - 1] >= S[p - 3] and S[p] >= S[p - 2]) \
                or\
                    (S[p - 2] < S[p - 3] and S[p - 1] <= S[p - 3] and S[p] <= S[p - 2]):
                # Extract two intermediate half cycles
                ampl = abs(S[p - 2] - S[p - 1])
                mean = (S[p - 2] + S[p - 1]) / 2
                ampl_mean.append((ampl, mean))
                ampl_mean.append((ampl, mean))

                S[p - 2] = S[p]

                p -= 2
            else:
                break

        if f == 0:
            pass
        else:
            break
    # phase 2
    q = 0
    while True:
        q += 1
        if p == q:
            break
        else:
            ampl = abs(S[q + 1] - S[q])
            mean = (S[q + 1] + S[q]) / 2
            ampl_mean.append((ampl, mean))
    return ampl_mean



'''
Created on 04/04/2016

@author: MMPE
'''
import glob
import multiprocessing
import os
import re
import threading

from wetb.utils.cluster_tools import pbswrap


def unix_path(path, cwd=None, fail_on_missing=False):
    """Convert case insensitive filename into unix case sensitive filename

    If no matching file or folder is found an error is raised

    Parameters
    ---------
    x : str
        Case insensitive filename or folder

    Returns
    -------
    Filename or folder name

    """
    if cwd:
        path = os.path.join(cwd, path)
    path = os.path.abspath(path)
    r = glob.glob(re.sub(r'([^:/\\])(?=[/\\]|$)', r'[\1]', path))
    if r:
        path = r[0]
    elif fail_on_missing:
        raise FileExistsError("File or folder matching '%s' not found"%path)
    if cwd:
        path = os.path.relpath(path, cwd)
    if os.path.isdir(path):
        path+="/"
    return path.replace("\\","/")

def unix_path_old(filename):
    filename = os.path.realpath(filename.replace("\\", "/")).replace("\\", "/")
    ufn, rest = os.path.splitdrive(filename)
    ufn += "/"
    for f in rest[1:].split("/"):
        f_lst = [f_ for f_ in os.listdir(ufn) if f_.lower() == f.lower()]
        if len(f_lst) > 1:
            f_lst = [f_ for f_ in f_lst if f_ == f]
        elif len(f_lst) == 0:
            raise IOError("'%s' not found in '%s'" % (f, ufn))
        else: # one match found
            ufn = os.path.join(ufn, f_lst[0])
    return ufn.replace("\\", "/")




class Resource(object):

    def __init__(self, min_cpu, min_free):
        self.min_cpu = min_cpu
        self.min_free = min_free
        self.cpu_free=0
        self.acquired = 0
        self.no_cpu="?"
        self.used_by_user=0
        self.resource_lock = threading.Lock()

    def ok2submit(self):
        """Always ok to have min_cpu cpus and ok to have more if there are min_free free cpus"""
        try:
            #print ("ok2submit")
            total, free, user = self.check_resources()
            user = max(user, self.acquired)
            user_max = max((self.min_cpu-user), self.cpu_free - self.min_free)
            #print ("ok2submit", total, free, user, user_max)
            return user_max
        except:
            return False
#         if user < self.min_cpu:
#             return True
#         elif free > self.min_free:
#             return True
#         else:
#             return False

    def acquire(self):
        with self.resource_lock:
            self.acquired += 1

    def release(self):
        with self.resource_lock:
            self.acquired -= 1


    def update_resource_status(self):
        try:
            self.no_cpu, self.cpu_free, self.used_by_user = self.check_resources()
        except Exception:
            pass


class SSHPBSClusterResource(Resource):
    finished = []
    loglines = {}
    is_executing = []
    
    def __init__(self, sshclient, min_cpu, min_free):
        Resource.__init__(self, min_cpu, min_free)
        self.ssh = sshclient
        self.resource_lock = threading.Lock()
        
    def glob(self, filepattern, cwd="", recursive=False):
        return self.ssh.glob(filepattern, cwd, recursive)

    @property
    def host(self):
        return self.ssh.host
    
    @property
    def username(self):
        return self.ssh.username


    def new_ssh_connection(self):
        from wetb.utils.cluster_tools.ssh_client import SSHClient
        return SSHClient(self.host, self.ssh.username, self.ssh.password, self.ssh.port)
        #return self.ssh

    def check_resources(self):
        with self.resource_lock:
            try:
                with self.ssh:
                    _, output, _ = self.ssh.execute('pbsnodes -l all')
                    pbsnodes, nodes = pbswrap.parse_pbsnode_lall(output.split("\n"))

                    _, output, _ = self.ssh.execute('qstat -n1')
                    users, host, nodesload = pbswrap.parse_qstat_n1(output.split("\n"), self.ssh.host)

                # if the user does not have any jobs, this will not exist
                try:
                    cpu_user = users[self.ssh.username]['cpus']
                    cpu_user += users[self.ssh.username]['Q']
                except KeyError:
                    cpu_user = 0
                cpu_user = max(cpu_user, self.acquired)
                cpu_free, nodeSum = pbswrap.count_cpus(users, host, pbsnodes)

                return nodeSum['used_cpu'] + cpu_free, cpu_free, cpu_user
            except Exception as e:
                raise EnvironmentError(str(e))


    def jobids(self, jobname_prefix):
            _, output, _ = self.ssh.execute('qstat -u %s' % self.username)
            return [l.split()[0].split(".")[0] for l in output.split("\n")[5:] if l.strip() != "" and l.split()[3].startswith("h2l")]

    def stop_pbsjobs(self, jobids):
        if not hasattr(jobids, "len"):
            jobids = list(jobids)
        self.ssh.execute("qdel %s" % (" ".join(jobids)))
        
    def setup_wine(self):    
        self.ssh.execute("""rm -f ./config-wine-hawc2.sh &&
wget https://gitlab.windenergy.dtu.dk/toolbox/pbsutils/raw/master/config-wine-hawc2.sh &&
chmod 777 config-wine-hawc2.sh &&
./config-wine-hawc2.sh""")
   


class LocalResource(Resource):
    def __init__(self, cpu_limit):

        Resource.__init__(self, cpu_limit, multiprocessing.cpu_count())
        #self.process_name = process_name
        self.host = 'Localhost'

    
    
    def check_resources(self):
        import psutil
        no_cpu = multiprocessing.cpu_count()
        cpu_free = (1 - psutil.cpu_percent(.1) / 100) * no_cpu
        used = self.acquired
        return no_cpu, cpu_free, used



# -*- coding: utf-8 -*-
"""
Created on Wed Nov  5 14:01:25 2014

@author: dave
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import str
from future.utils import viewitems
from future import standard_library
standard_library.install_aliases()

import os
import unittest
from glob import glob

import pandas as pd

from wetb.prepost import misc
from wetb.prepost.GenerateHydro import hydro_input
from wetb.prepost import hawcstab2


def casedict2xlsx():
    """
    Convert a full Cases.cases dict to Excel spreadsheets
    """


def configure_dirs(verbose=False, pattern_master='*_master_*'):
    """
    Automatically configure required directories to launch simulations
    """

    P_RUN = str(os.getcwd())
    p_run_root = os.sep.join(P_RUN.split(os.sep)[:-2])
    # MODEL SOURCES, exchanche file sources
    P_SOURCE = P_RUN
    # Project name, sim_id: derive from folder name
    PROJECT = P_RUN.split(os.sep)[-2]
    sim_id = P_RUN.split(os.sep)[-1]

    master = find_master_file(P_SOURCE, pattern=pattern_master)
    if master is None:
        raise ValueError('Could not find master file in htc/_master')
    MASTERFILE = master
    P_MASTERFILE = os.path.join(P_SOURCE, 'htc%s_master%s' % (os.sep, os.sep))
    POST_DIR = os.path.join(p_run_root, PROJECT, sim_id, 'prepost-data%s' % os.sep)

    if verbose:
        print('='*79)
        print('POST_DIR: %s' % POST_DIR)
        print('   P_RUN: %s' % P_RUN)
        print('P_SOURCE: %s' % P_SOURCE)
        print(' PROJECT: %s' % PROJECT)
        print('  sim_id: %s' % sim_id)
        print('  master: %s' % MASTERFILE)
        print('='*79)

    return P_RUN, P_SOURCE, PROJECT, sim_id, P_MASTERFILE, MASTERFILE, POST_DIR


def find_master_file(proot, htc_dir='htc', master_dir='_master',
                     pattern='*_master_*'):
    """
    Find the master file name. It is assumed that the master file is in the
    folder _master, under htc, and contains _master_ in the file name. If
    multiple files contain pattern, the last file of the sorted list is
    returned.

    Parameters
    ----------

    proot

    htc_dir : str, default: htc

    master_dir : str, default: _master

    pattern : str, default: *_master_*

    """

    fpath_search = os.path.join(proot, htc_dir, master_dir, pattern)
    files = glob(fpath_search)
    if len(files) > 0:
        return sorted(files)[-1]
    return None


def variable_tag_func(master, case_id_short=False):
    """
    When using the Excel definitions, and the whole default setup, the
    variable_tag_func is not required to do anything extra.
    """

    # -------------------------------------------------------------------------
#    mt = master
#    V = mt['windspeed']
#    mt['duration'] = mt['time_stop'] - mt['t0']
#    t = mt['duration']
#    if V > abs(1e-15):
#        b = 5.6
#        mt['TI'] = mt['TI_ref'] * ((0.75*V) + b) / V # NTM
#        # ETM
#        c = 2.0
#        V_ave = 0.2 * 50.0
#        sigma = mt['TI_ref'] / V
#        mt['TI'] = sigma * c * (0.072 * (V_ave / c + 3.0) * (V / c - 4.0) + 10.0)
#    else:
#        mt['TI'] = 0
#
#    mt['turb_dx'] = V*t/mt['turb_grid_x']
#
#    mt['turb_dy'] = (mt['rotor_diameter'] / mt['turb_grid_yz'])*1.1
#
#    mt['turb_dz'] = (mt['rotor_diameter'] / mt['turb_grid_yz'])*1.1
#
#    # check: dx spacing should be 0.1*mean_windspeed and 0.2*mean_windspeed
#    # between 0.1 and 0.2 seconds between points
#    if not (V*0.1 < mt['turb_dx'] < V*0.2):
#        logging.warn('turbulence spacing dx out of bounds')
#        print('%5.3f  %5.3f  %5.3f' % (V*0.1, mt['turb_dx'], V*0.2))
#
#    #mt['turb_base_name'] = 'turb_s' + str(mt['turb_seed']) + '_' + str(V)
#    mt['turb_base_name'] = 'turb_s%i_%1.2f' % (mt['turb_seed'], V)
    # -------------------------------------------------------------------------

    return master


def vartag_dlcs(master):

    mt = master.tags

    dlc_case = mt['[Case folder]']
    mt['[data_dir]'] = 'data/'
    mt['[res_dir]'] = 'res/%s/' % dlc_case
    mt['[log_dir]'] = 'logfiles/%s/' % dlc_case
    mt['[htc_dir]'] = 'htc/%s/' % dlc_case
    mt['[case_id]'] = mt['[Case id.]']
    mt['[time_stop]'] = mt['[time stop]']
    mt['[turb_base_name]'] = mt['[Turb base name]']
    mt['[DLC]'] = mt['[Case id.]'].split('_')[0][3:]
    mt['[pbs_out_dir]'] = 'pbs_out/%s/' % dlc_case
    mt['[pbs_in_dir]'] = 'pbs_in/%s/' % dlc_case
    mt['[iter_dir]'] = 'iter/%s/' % dlc_case
    if mt['[eigen_analysis]']:
        rpl = (dlc_case, mt['[Case id.]'])
        mt['[eigenfreq_dir]'] = 'res_eigen/%s/%s/' % rpl
    mt['[duration]'] = str(float(mt['[time_stop]']) - float(mt['[t0]']))
    # replace nan with empty
    for ii, jj in mt.items():
        if jj == 'nan':
            mt[ii] = ''

    return master


def vartag_excel_stabcon(master):
    """Variable tag function type that generates a hydro input file for the
    wave kinematics dll if [hydro input name] is defined properly.
    """

    mt = master.tags
    if '[hydro input name]' not in mt or not mt['[hydro input name]']:
        return master

    print('creating hydro input file for: %s.inp\n' % mt['[hydro input name]'])

    mt['[wdepth]'] = float(mt['[wdepth]'])
    mt['[Hs]'] = float(mt['[Hs]'])
    mt['[Tp]'] = float(mt['[Tp]'])


    if '[wave_gamma]' not in mt or not mt['[wave_gamma]']:
        mt['[wave_gamma]'] = 3.3
    else:
        mt['[wave_gamma]'] = float(mt['[wave_gamma]'])

    if '[wave_coef]' not in mt or not mt['[wave_coef]']:
        mt['[wave_coef]'] = 200
    else:
        mt['[wave_coef]'] = int(mt['[wave_coef]'])

    if '[stretching]' not in mt or not mt['[stretching]']:
        mt['[stretching]'] = 1
    else:
        mt['[stretching]'] = int(mt['[stretching]'])

    if '[wave_seed]' not in mt or not mt['[wave_seed]']:
        mt['[wave_seed]'] = int(mt['[seed]'])
    else:
        mt['[wave_seed]'] = int(mt['[wave_seed]'])

    try:
        embed_sf = float(master.tags['[embed_sf]'])
        embed_sf_t0 = int(master.tags['[t0]']) + 20
    except KeyError:
        embed_sf = None
        embed_sf_t0 = None

    hio = hydro_input(wavetype=mt['[wave_type]'], Hs=mt['[Hs]'], Tp=mt['[Tp]'],
                      gamma=mt['[wave_gamma]'], wdepth=mt['[wdepth]'],
                      spectrum=mt['[wave_spectrum]'], seed=mt['[wave_seed]'],
                      stretching=mt['[stretching]'], coef=mt['[wave_coef]'],
                      embed_sf=embed_sf, embed_sf_t0=embed_sf_t0, spreading=None)

    hio.execute(filename=mt['[hydro input name]'] + '.inp',
                folder=mt['[hydro_dir]'])

    return master


def tags_dlcs(master):
    """
    Initiate tags that are defined in the DLC spreadsheets
    """

    master.tags['[t0]'] = 0
    master.tags['[time stop]'] = 0
    master.tags['[Case folder]'] = 'test'
    master.tags['[Case id.]'] = 'test'
    master.tags['[Windspeed]'] = 8
    master.tags['[wdir]'] = 0 # used for the user defined wind
    master.tags['[wdir_rot]'] = 0 # used for the windfield rotations
    master.tags['[seed]'] = None
    master.tags['[tu_model]'] = 0
    master.tags['[TI]'] = 0
    master.tags['[Turb base name]'] = 'none'
    master.tags['[turb_dx]'] = 1.0
    master.tags['[shear_exp]'] = 0.2
    master.tags['[wsp factor]'] = 1.0
    master.tags['[gust]'] = False
    master.tags['[gust_type]'] = ''
    master.tags['[G_A]'] = ''
    master.tags['[G_phi0]'] = ''
    master.tags['[G_t0]'] = ''
    master.tags['[G_T]'] = ''
    master.tags['[Rotor azimuth]'] = 0
    master.tags['[Free shaft rot]'] = ''
    master.tags['[init_wr]'] = 0.5
    master.tags['[Pitch 1 DLC22b]'] = 0
    master.tags['[Rotor locked]'] = False
    master.tags['[Time stuck DLC22b]'] = -1
    master.tags['[Cut-in time]'] = -1
    master.tags['[Cut-out time]'] = -1
    master.tags['[Stop type]'] = -1
    master.tags['[Pitvel 1]'] = 4
    master.tags['[Pitvel 2]'] = 6
    master.tags['[Grid loss time]'] = 1000
    master.tags['[out_format]'] = 'hawc_binary'
    master.tags['[Time pitch runaway]'] = 1000
    master.tags['[Induction]'] = 1
    master.tags['[Dyn stall]'] = 1
    # required tags for the MannTurb64 standalone turbulence box generator
    master.tags['[MannAlfaEpsilon]'] = 1.0
    master.tags['[MannL]'] = 29.4
    master.tags['[MannGamma]'] = 3.0
    master.tags['[turb_nr_u]'] = 8192
    master.tags['[turb_nr_v]'] = 32
    master.tags['[turb_nr_w]'] = 32
    master.tags['[turb_dx]'] = 1
    master.tags['[turb_dy]'] = 6.5
    master.tags['[turb_dz]'] = 6.5
    master.tags['[high_freq_comp]'] = 1

    return master


def tags_defaults(master):

    # other required tags and their defaults
    master.tags['[dt_sim]'] = 0.02
    master.tags['[hawc2_exe]'] = 'hawc2-latest'
    # folder names for the saved results, htc, data, zip files
    # Following dirs are relative to the model_dir_server and they specify
    # the location of where the results, logfiles, animation files that where
    # run on the server should be copied to after the simulation has finished.
    # on the node, it will try to copy the turbulence files from these dirs
    master.tags['[animation_dir]'] = 'animation/'
    master.tags['[control_dir]']   = 'control/'
    master.tags['[data_dir]']      = 'data/'
    master.tags['[eigen_analysis]'] = False
    master.tags['[eigenfreq_dir]'] = False
    master.tags['[htc_dir]']       = 'htc/'
    master.tags['[log_dir]']       = 'logfiles/'
    master.tags['[meander_dir]']   = False
    master.tags['[opt_dir]']       = False
    master.tags['[pbs_in_dir]']    = 'pbs_in/'
    master.tags['[pbs_out_dir]']   = 'pbs_out/'
    master.tags['[res_dir]']       = 'res/'
    master.tags['[iter_dir]']      = 'iter/'
    master.tags['[turb_dir]']      = 'turb/'
    master.tags['[turb_db_dir]']   = '../turb/'
    master.tags['[micro_dir]']      = False
    master.tags['[hydro_dir]']     = False
    master.tags['[mooring_dir]']   = False
    master.tags['[externalforce]'] = False
    # required tags for the MannTurb64 standalone turbulence box generator
    master.tags['[MannAlfaEpsilon]'] = 1.0
    master.tags['[MannL]'] = 29.4
    master.tags['[MannGamma]'] = 3.0
    master.tags['[turb_nr_u]'] = 8192
    master.tags['[turb_nr_v]'] = 32
    master.tags['[turb_nr_w]'] = 32
    master.tags['[turb_dx]'] = 1
    master.tags['[turb_dy]'] = 6.5
    master.tags['[turb_dz]'] = 6.5
    master.tags['[high_freq_comp]'] = 1
    # zip_root_files only is used when copy to run_dir and zip creation, define
    # in the HtcMaster object
    master.tags['[zip_root_files]'] = []
    # only active on PBS level, so files have to be present in the run_dir
    master.tags['[copyback_files]'] = []   # copyback_resultfile
    master.tags['[copyback_frename]'] = [] # copyback_resultrename
    master.tags['[copyto_files]'] = []     # copyto_inputfile
    master.tags['[copyto_generic]'] = []   # copyto_input_required_defaultname
    master.tags['[eigen_analysis]'] = False

    # =========================================================================
    # basic required tags by HtcMaster and PBS in order to function properly
    # =========================================================================
    # the express queue ('#PBS -q xpresq') has a maximum walltime of 1h
    master.tags['[pbs_queue_command]'] = '#PBS -q workq'
    # walltime should have following format: hh:mm:ss
    master.tags['[walltime]'] = '04:00:00'
    master.tags['[auto_walltime]'] = False

    return master


def excel_stabcon(proot, fext='xlsx', pignore=None, pinclude=None, sheet=0,
                  silent=False, p_source=False):
    """
    Read all MS Excel files that hold load case definitions according to
    the team STABCON definitions. Save each case in a list according to the
    opt_tags principles as used in Simulations.launch(). This method assumes
    that a standard HAWC2 folder layout is used with the following folder
    names: res, logfiles, htc, pbs_out, pbs_in, iter. Further some tags
    are added to be compatible with the tag convention in the Simulations
    module.

    The opt_tags case list is sorted according to the Excel file names, and
    follows the same ordering as in each of the different Excel files.


    Parameters
    ----------

    proot : string
        Path that will be searched recursively for Excel files containing
        load case definitions.

    fext : string, default='xlsx'
        File extension of the Excel files that should be loaded

    pignore : string, default=None
        Specify which string can not occur in the full path of the DLC target.

    pinclude : string, default=None
        Specify which string has to occur in the full path of the DLC target.

    sheet : string or int, default=0
        Name or index of the Excel sheet to be considered. By default, the
        first sheet (index=0) is taken.

    p_source : string, default=False


    Returns
    -------

    opt_tags : list of dicts
        A list of case dictionaries, where each case dictionary holds all
        the tag/value key pairs for a single given case.

    """
    if not silent:
        print('looking for DLC spreadsheet definitions at:')
        print(proot)
    dict_dfs = misc.read_excel_files(proot, fext=fext, pignore=pignore,
                                    sheet=sheet, pinclude=pinclude,
                                    silent=silent)

    if not silent:
        print('found %i Excel file(s), ' % len(dict_dfs), end='')

    if not silent:
        k = 0
        for df in dict_dfs:
            k += len(df)
        print('in which a total of %i cases are defined.' % k)

    opt_tags = []

    for (dlc, df) in sorted(viewitems(dict_dfs)):
        # replace ';' with False, and Nan(='') with True
        # this is more easy when testing for the presence of stuff compared
        # to checking if a value is either True/False or ''/';'
        # this doesn't work, it will result in 1 for True and 0 for False
        # because the nan values have np.float dtype
#        df.fillna(' ', inplace=True)
#        df.replace(';', False, inplace=True)
        # instead, convert everything to strings, this will maintain some nans
        # as empty strings, but not all of them!
        df2 = df.astype(str)
        for count, row in df2.iterrows():
            tags_dict = {}
            # construct to dict, convert unicode keys/values to strings
            for key, value in row.iteritems():
                if isinstance(value, str):
                    tags_dict[str(key)] = str(value)
                else:
                    tags_dict[str(key)] = value
                # convert ; and empty to False/True, "none" to None
                if isinstance(tags_dict[str(key)], str):
                    if tags_dict[str(key)] == ';':
                        tags_dict[str(key)] = False
                    elif tags_dict[str(key)] == '':
                        tags_dict[str(key)] = True
                    elif tags_dict[str(key)].lower() == 'nan':
                        tags_dict[str(key)] = True
                    elif tags_dict[str(key)].lower() == 'none':
                        tags_dict[str(key)] = None

            # FIXME: this horrible mess requires a nice and clearly defined
            # tag spec/naming convention, and with special tag prefix
            if '[Windspeed]' not in tags_dict and '[wsp]' in tags_dict:
                tags_dict['[Windspeed]'] = tags_dict['[wsp]']
            elif '[Windspeed]' in tags_dict and '[wsp]' not in tags_dict:
                tags_dict['[wsp]'] = tags_dict['[Windspeed]']
            # avoid that any possible default tags from wetb will be used
            # instead of the ones from the spreadsheet
            if '[seed]' in tags_dict:
                tags_dict['[tu_seed]'] = tags_dict['[seed]']
            # in case people are using other turbulence tag names in the sheet
            elif '[tu_seed]' in tags_dict:
                tags_dict['[seed]'] = tags_dict['[tu_seed]']
            elif '[turb_seed]' in tags_dict:
                tags_dict['[seed]'] = tags_dict['[turb_seed]']
            else:
                raise KeyError('[seed] should be used as tag for turb. seed')

            # for backwards compatibility
            if '[wake_dir]' in tags_dict and not '[micro_dir]':
                tags_dict['[micro_dir]'] = tags_dict['[wake_dir]']
            if '[wake_db_dir]' in tags_dict and not '[micro_db_dir]':
                tags_dict['[micro_db_dir]'] = tags_dict['[wake_db_dir]']
            if '[wake_base_name]' in tags_dict and not '[micro_base_name]':
                tags_dict['[micro_base_name]'] = tags_dict['[wake_base_name]']

            tags_dict['[Case folder]'] = tags_dict['[Case folder]'].lower()
            tags_dict['[Case id.]'] = tags_dict['[Case id.]'].lower()
            dlc_case = tags_dict['[Case folder]']
            if '[data_dir]' not in tags_dict:
                tags_dict['[data_dir]'] = 'data/'
            if '[res_dir]' not in tags_dict:
                tags_dict['[res_dir]'] = 'res/%s/' % dlc_case
            if '[log_dir]' not in tags_dict:
                tags_dict['[log_dir]'] = 'logfiles/%s/' % dlc_case
            if '[htc_dir]' not in tags_dict:
                tags_dict['[htc_dir]'] = 'htc/%s/' % dlc_case
            if '[Case id.]' in tags_dict.keys():
                tags_dict['[case_id]'] = tags_dict['[Case id.]']
            if '[time stop]' in tags_dict.keys():
                tags_dict['[time_stop]'] = tags_dict['[time stop]']
            else:
                tags_dict['[time stop]'] = tags_dict['[time_stop]']

            if '[Turb base name]' in tags_dict:
                if not '[turb_base_name]' in tags_dict:
                    tags_dict['[turb_base_name]'] = tags_dict['[Turb base name]']
            elif not '[turb_base_name]' in tags_dict:
                tags_dict['[turb_base_name]'] = None
                tags_dict['[Turb base name]'] = None
            # for backwards compatibility: user has only defined [Turb base name]
            elif not '[Turb base name]' in tags_dict:
                tags_dict['[Turb base name]'] = tags_dict['[turb_base_name]']

            tags_dict['[DLC]'] = tags_dict['[Case id.]'].split('_')[0][3:]
            if '[pbs_out_dir]' not in tags_dict:
                tags_dict['[pbs_out_dir]'] = 'pbs_out/%s/' % dlc_case
            if '[pbs_in_dir]' not in tags_dict:
                tags_dict['[pbs_in_dir]'] = 'pbs_in/%s/' % dlc_case
            if '[iter_dir]' not in tags_dict:
                tags_dict['[iter_dir]'] = 'iter/%s/' % dlc_case
            # the default spreadsheets do not define the tags related to the
            # eigen analsyis yet
            if '[eigen_analysis]' in tags_dict and tags_dict['[eigen_analysis]']:
                rpl = (dlc_case, tags_dict['[Case id.]'])
                if '[eigenfreq_dir]' in tags_dict:
                    tags_dict['[eigenfreq_dir]'] = 'res_eigen/%s/%s/' % rpl
            t_stop = float(tags_dict['[time_stop]'])
            t0 = float(tags_dict['[t0]'])
            tags_dict['[duration]'] = str(t_stop - t0)
            # in case there is a controller input file defined
            if '[controller_tuning_file]' in tags_dict:
                hs2 = hawcstab2.ReadControlTuning()
                # absolute path of the model root containing the tuning file
                fpath = tags_dict['[controller_tuning_file]']
                if p_source:
                    fpath = os.path.join(p_source, fpath)
                hs2.read_parameters(fpath)
                tags_dict['[pi_gen_reg1.K]'] = hs2.pi_gen_reg1.K
                tags_dict['[pi_gen_reg2.Kp]'] = hs2.pi_gen_reg2.Kp
                tags_dict['[pi_gen_reg2.Ki]'] = hs2.pi_gen_reg2.Ki
                tags_dict['[pi_gen_reg2.Kd]'] = hs2.pi_gen_reg2.Kd
                tags_dict['[pi_pitch_reg3.Kp]'] = hs2.pi_pitch_reg3.Kp
                tags_dict['[pi_pitch_reg3.Ki]'] = hs2.pi_pitch_reg3.Ki
                tags_dict['[pi_pitch_reg3.Kd]'] = hs2.pi_pitch_reg3.Kd
                tags_dict['[pi_pitch_reg3.K1]'] = hs2.pi_pitch_reg3.K1
                tags_dict['[pi_pitch_reg3.K2]'] = hs2.pi_pitch_reg3.K2
                tags_dict['[aero_damp.Kp2]'] = hs2.aero_damp.Kp2
                tags_dict['[aero_damp.Ko1]'] = hs2.aero_damp.Ko1
                tags_dict['[aero_damp.Ko2]'] = hs2.aero_damp.Ko2
            # save a copy of the current case an one opt_tags entry
            opt_tags.append(tags_dict.copy())

    return opt_tags


def read_tags_spreadsheet(fname):
    """Read a spreadsheet with HAWC2 tags, make sure no 0/1/nan ends up
    replacing the ";" or "" (empty). Do not add any other tags.

    Returns
    -------

    opt_tags : [{}, {}] list of dictionaries
    """

    df = pd.read_excel(fname)
    df2 = df.astype(str)
    opt_tags = []
    for count, row in df2.iterrows():
        tags_dict = {}
        # construct to dict, convert unicode keys/values to strings
        for key, value in row.items():
            if isinstance(value, str):
                tags_dict[str(key)] = str(value)
            else:
                tags_dict[str(key)] = value
            # convert ; and empty to False/True
            if tags_dict[str(key)] == ';':
                tags_dict[str(key)] = False
            elif tags_dict[str(key)] == '':
                tags_dict[str(key)] = True
            elif tags_dict[str(key)].lower() == 'nan':
                tags_dict[str(key)] = True
        opt_tags.append(tags_dict.copy())

    return opt_tags


class Tests(unittest.TestCase):
    """
    """

    def setUp(self):
        self.fpath = os.path.join(os.path.dirname(__file__), 'data/DLCs')

    def test_read_tag_exchange_file(self):

        df_list = misc.read_excel_files(self.fpath, fext='xlsx', pignore=None,
                                        sheet=0, pinclude=None)

#        df = df_list[list(df_list.keys())[0]]
#        df.fillna('', inplace=True)
#        df.replace(';', False, inplace=True)

    def test_excel_stabcon(self):
        opt_tags = excel_stabcon(self.fpath)


if __name__ == '__main__':

    unittest.main()



'''
Created on 24/04/2014

@author: MMPE
'''
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from io import open
from future import standard_library
from wetb.hawc2.htc_file import HTCFile
standard_library.install_aliases()
import numpy as np
import os


class ShearFile(object):
    """HAWC2 user defined shear file
    
    Examples:
    ---------
    
    >>> sf = ShearFile([-55, 55], [30, 100, 160] , u=np.array([[0.7, 1, 1.3], [0.7, 1, 1.3]]).T)
    >>> print (sf.uvw([-55,0],[65,135])) #uvw factors
    [array([ 0.85 ,  1.175]), array([ 0.,  0.]), array([ 0.,  0.])]
    >>> from wetb.wind.shear import power_shear
    >>> print (sf.uvw([-55,0],[65,135], shear=power_shear(.2,100,10))) # uvw wind speeds
    [array([  7.79832978,  12.47684042]), array([ 0.,  0.]), array([ 0.,  0.])]
    >>> sf.save('test.dat')
    """
    
    def __init__(self,v_positions, w_positions, u=None, v=None, w=None, u0=None, shear=None):
        """
        Parameters
        ----------
        v_positions : array_like
            lateral coordinates
        w_positions : array_like
            vertical coordinates
        u : array_like, optional
            shear_u component, normalized with U_mean\n
            shape must be (#w_positions, #v_positions) or (#w_positions,)
        v : array_like, optional
            shear_v component, normalized with U_mean\n
            shape must be (#w_positions, #v_positions) or (#w_positions,)
        w : array_like, optional
            shear_w component, normalized with U_mean\n
            shape must be (#w_positions, #v_positions) or (#w_positions,)
        u0 : int or float
            Mean wind speed
        shear : function
            Shear function: f(height)->wsp
        """
        self.v_positions = v_positions
        self.w_positions = w_positions
        shape = (len(w_positions), len(v_positions))
        uvw = [u, v, w]
        for i in range(3):
            if uvw[i] is None:
                if i == 0:
                    uvw[i] = np.ones((shape))
                else:
                    uvw[i] = np.zeros((shape))
            else:
                uvw[i] = np.array(uvw[i])
                if len(uvw[i].shape) == 1 and uvw[i].shape[0] == shape[0]:
                    uvw[i] = np.repeat(np.atleast_2d(uvw[i]).T, shape[1], 1)
    
                assert uvw[i].shape == shape, (i, uvw[i].shape, shape)
        self.u, self.v, self.w = uvw
        self.u0 = u0
        self.shear = shear
    
    def uvw(self, v,w,u0=None,shear=None):
        """Calculate u,v,w wind speeds at position(s) (v,w)
        Parameters
        ----------
        v : int, float or array_like
            v-coordinate(s)
        w : int, float or array_like
            w-coordinates(s)
        u0 : int or float
            mean wind speed
        shear : function or None
            if function: f(height)->wsp
            if None: self.shear is used, if not None. 
            Otherwise wind speed factors instead of absolute wind speeds are returned
            
        Returns
        -------
        u,v,w 
            wind speed(s) or wind speed factor(s) if shear not defined
        """
        u0 =u0 or self.u0 or 1
        shear = shear or self.shear or (lambda z: 0)
        from scipy.interpolate import RegularGridInterpolator
        wv = np.array([w,v]).T
        return [(RegularGridInterpolator((self.w_positions, self.v_positions), uvw)(wv))*u0+shear(w) 
                            for uvw in [self.u, self.v, self.w] if uvw is not None]
        


    def save(self, filename):
        """Save user defined shear file
        Parameters
        ----------
        filename : str:
            Filename
        """
        
        # exist_ok does not exist in Python27
        filename = os.path.abspath(filename)
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))#, exist_ok=True)
        with open(filename, 'w', encoding='utf-8') as fid:
            fid.write(" # autogenerated shear file\n")
            fid.write("  %d %d\n" % (len(self.v_positions), len(self.w_positions)))
            
            for i, (l,vuw) in enumerate(zip(['v', 'u', 'w'],[self.v,self.u,self.w])):
                fid.write(" # shear %s component\n  " % l)
                fid.write("\n  ".join([" ".join(["%.10f" % v for v in r ]) for r in vuw]))
                fid.write("\n")
            for yz, coor in (['v', self.v_positions], ['w', self.w_positions]):
                fid.write(" # %s coordinates\n  " % yz)
                fid.write("\n  ".join("%.10f" % v for v in coor))
                fid.write("\n")

    @staticmethod
    def load(filename):
        """Load shear file
        Parameters 
        ----------
        filename : str
            Filename
            
        Returns
        -------
        shear file : ShearFile-object
        """
        with open(filename) as fid:
            lines = fid.readlines()
        no_V,no_W=map(int,lines[1].split())
        v,u,w = [np.array([row.split() for row in lines[3+(no_W+1)*i:3+(no_W+1)*(i+1)-1] ],dtype=float) for i in range(3)]
        v_positions = np.array(lines[3+(no_W+1)*3:3+(no_W+1)*3+no_V],dtype=float)
        w_positions = np.array(lines[3+(no_W+1)*3+(no_V+1):3+(no_W+1)*3+(no_V+1)+no_W],dtype=float)
        return ShearFile(v_positions,w_positions,u,v,w)

    @staticmethod
    def load_from_htc(htc_file):
        """Load shear file from HTC file including shear function
        Parameters 
        ----------
        htc_file : str or HTCFile
            Filename or HTCFile
            
        Returns
        -------
        shear file : ShearFile-object
        """
        if isinstance(htc_file,str):
            htc_file = HTCFile(htc_file)
        user_defined_shear_filename = os.path.join(htc_file.modelpath, htc_file.wind.user_defined_shear[0])
        shear_file = ShearFile.load(user_defined_shear_filename)
        shear_file.shear = htc_file.get_shear()
        shear_file.u0 = htc_file.wind.wsp[0]
        return shear_file

def save(filename, v_coordinates, w_coordinates, u=None, v=None, w=None):
    """Save shear file (deprecated)"""
    ShearFile(v_coordinates, w_coordinates,u,v,w).save(filename)





if __name__ == "__main__":
    from wetb.wind.shear import power_shear
    sf = ShearFile([-55, 55], [30, 100, 160] , u=np.array([[0.7, 1, 1.3], [0.7, 1, 1.3]]).T)
    print (sf.uvw([-55,0],[65,135])) #uvw factors
    #[array([ 0.85 ,  1.175]), array([ 0.,  0.]), array([ 0.,  0.])]
    print (sf.uvw([-55,0],[65,135], shear=power_shear(.2,100,10))) # uvw wind speeds
    #[array([  7.79832978,  12.47684042]), array([ 0.,  0.]), array([ 0.,  0.])]
    sf.save('test.dat')
    



from wetb.dlb import iec64100_1


import pytest
from wetb.dlb.hawc2_iec_dlc_writer import HAWC2_IEC_DLC_Writer

import os
from wetb.hawc2.tests import test_files
import shutil
from tests import npt, run_main
from wetb.hawc2.htc_file import HTCFile
from tests.run_main import run_module_main
from wetb.dlb.iec64100_1 import DTU_IEC64100_1_Ref_DLB

path = os.path.dirname(test_files.__file__) + '/simulation_setup/DTU10MWRef6.0/'


def test_hawc2_writer_main():
    run_module_main(iec64100_1)


def clean_up():
    for f in [f for f in os.listdir(path + 'htc') if f.lower().startswith('dlc')]:
        if os.path.isdir(path + 'htc/' + f):
            shutil.rmtree(path + 'htc/' + f)


@pytest.yield_fixture(autouse=True)
def run_around_tests():
    clean_up()
    yield
    clean_up()


@pytest.fixture
def writer():
    return HAWC2_IEC_DLC_Writer(path + 'htc/DTU_10MW_RWT.htc', diameter=127)


def test_main():
    run_main.run_module_main(iec64100_1)


def test_DLC12(writer):
    dlc12 = DTU_IEC64100_1_Ref_DLB(iec_wt_class='1A', Vin=4, Vout=26, Vr=10, D=180, z_hub=90)['DLC12']
    assert len(dlc12) == 216  # 12 wsp, 3 wdir, 6 seeds
    writer.from_pandas(dlc12[::24][:2])
    npt.assert_array_equal(sorted(os.listdir(path + "htc/DLC12")),
                           ['DLC12_wsp04_wdir350_s1001.htc', 'DLC12_wsp06_wdir000_s1101.htc'])
    htc = HTCFile(path + "htc/dlc12/dlc12_wsp04_wdir350_s1001.htc")
    assert htc.wind.wsp[0] == 4
    npt.assert_array_equal(htc.wind.windfield_rotations.values, [-10, 0, 0])
    assert htc.wind.turb_format[0] == 1
    assert htc.wind.mann.create_turb_parameters[3] == 1001


def test_DLC21(writer):
    dlc = DTU_IEC64100_1_Ref_DLB(iec_wt_class='1A', Vin=4, Vout=26, Vr=10, D=180, z_hub=90)['DLC21']
    assert len(dlc) == 144  # 12 wsp, 3 wdir, 4 seeds
    writer.from_pandas(dlc[::16][:2])
    npt.assert_array_equal(sorted(os.listdir(path + "htc/DLC21")),
                           ['DLC21_wsp04_wdir350_s1001.htc', 'DLC21_wsp06_wdir000_s1101.htc'])
    htc = HTCFile(path + "htc/dlc21/dlc21_wsp04_wdir350_s1001.htc")
    assert htc.wind.wsp[0] == 4
    npt.assert_array_equal(htc.wind.windfield_rotations.values, [-10, 0, 0])
    assert htc.wind.turb_format[0] == 1
    assert htc.wind.mann.create_turb_parameters[3] == 1001
    assert htc.dll.get_subsection_by_name('generator_servo', 'name').init.constant__7.values == [7, 110]


def test_DLC22y(writer):
    dlc = DTU_IEC64100_1_Ref_DLB(iec_wt_class='1A', Vin=4, Vout=26, Vr=10, D=180, z_hub=90)['DLC22y']
    assert len(dlc) == 276  # 12 wsp, 23 wdir, 1 seeds
    writer.from_pandas(dlc[::24][:2])
    npt.assert_array_equal(sorted(os.listdir(path + "htc/DLC22y")),
                           ['DLC22y_wsp04_wdir015_s1001.htc', 'DLC22y_wsp06_wdir030_s1101.htc'])
    htc = HTCFile(path + "htc/dlc22y/dlc22y_wsp04_wdir015_s1001.htc")
    assert htc.wind.wsp[0] == 4
    npt.assert_array_equal(htc.wind.windfield_rotations.values, [15, 0, 0])
    assert htc.wind.turb_format[0] == 1
    assert htc.wind.mann.create_turb_parameters[3] == 1001



'''
Created on 13. jan. 2017

@author: mmpe
'''
import unittest

from scipy import signal

import numpy as np
from wetb.signal.filters import first_order
from wetb.signal.fit._fourier_fit import F2x, x2F, rx2F


class Test_first_order_filters(unittest.TestCase):


    def test_low_pass(self):
        a = np.random.randint(0,100,100).astype(np.float)
        b = first_order.low_pass(a, 1, 1)
        self.assertLess(b.std(), a.std())
        if 0:
            import matplotlib.pyplot as plt
            plt.plot(a)
            plt.plot(b)
            plt.show()
          
          
    def test_low_pass2(self):
        t = np.linspace(0, 1.0, 2001)
        xlow = np.sin(2 * np.pi * 5 * t)
        xhigh = np.sin(2 * np.pi * 250 * t)
        x = xlow + xhigh
                
        b, a = signal.butter(8, 0.125)
        y = signal.filtfilt(b, a, x, padlen=150)
        self.assertAlmostEqual(np.abs(y - xlow).max(),0,4)
        if 0:
            import matplotlib.pyplot as plt
            plt.plot(x)
            plt.plot(y)
            plt.show()
        

#     def test_low_pass3(self):
#         t = np.linspace(0, 1.0, 2001)
# #         xlow = np.sin(2 * np.pi * 5 * t)
# #         xhigh = np.sin(2 * np.pi * 250 * t)
# #         x = xlow + xhigh
#         x = np.sum([np.sin(t*x*2*np.pi) for x in range(1,200)],0)                
#         cutoff = .2
#         b, a = signal.butter(8,cutoff)
#         w, h = signal.freqs(b, a)
#         y = signal.filtfilt(b, a, x)
#         F = rx2F(x, max_nfft=len(t))
#         tF = np.linspace(0, len(F)/t[-1],len(F))
#         Fy = rx2F(y, max_nfft=len(t))
# 
#         if 1:
#             import matplotlib.pyplot as plt
# #             plt.plot(x)
# #             plt.plot(y)
# #             plt.show()
# 
#             plt.plot(tF, np.abs(F))
#             plt.plot(tF, np.abs(Fy))
#             plt.xlim([0,260])
#             plt.show()
#             print (b,a)
#             
#             plt.plot(w, 20 * np.log10(abs(h)))
#             plt.xscale('log')
#             plt.title('Butterworth filter frequency response')
#             plt.xlabel('Frequency [radians / second]')
#             plt.ylabel('Amplitude [dB]')
#             plt.margins(0, 0.1)
#             plt.grid(which='both', axis='both')
#             plt.axvline(cutoff, color='green') # cutoff frequency
#             plt.show()
#      
#     def test_low_pass2(self):
#         t = np.arange(100000)/10000
#         dt = t[1]-t[0]
#         hz2 = np.sin(t*2*2*np.pi) # frq = 2 hz or 2*2pi rad/s
#         hz10 = np.sin(t*10*2*np.pi) # frq = 10 hz or 10*2pi rad/s
#         hz20 = np.sin(t*100*2*np.pi) # frq = 100 hz or 100*2pi rad/s
#         #sig = np.sum([np.sin(t*x*2*np.pi) for x in range(1,200)],0)
#         sig = np.sum([np.sin(t*x*2*np.pi) for x in [1,5,10]],0)
#         print (sig.shape)
#         F = rx2F(sig, max_nfft=len(t))
#         tF = np.linspace(0, len(F)/t[-1],len(F))
#           
#           
#         sig_lp1 = first_order.low_pass(sig, dt, 30*dt)
#         F_lp1 = rx2F(sig_lp1, max_nfft=len(t))
#    
#         bb, ba = signal.butter(1,5, 'low', analog=True)
#         print (bb, ba)
#         w1, h1 = signal.freqs(bb, ba)
#         sig_lp2 = signal.filtfilt(bb,ba, sig)
# #         print (sig_lp2)
# #         F_lp2 = rx2F(sig_lp2, max_nfft=len(t))
# #         bb, ba = signal.butter(10,10, 'low', analog=True)
# #         sig_lp3 = signal.lfilter(bb,ba, sig)
# #         F_lp3 = rx2F(sig_lp3, max_nfft=len(t))
# #         w2, h2 = signal.freqs(bb, ba)
# # #          
# #         self.assertLess(b.std(), a.std())
#         if 1:
#             import matplotlib.pyplot as plt
#             plt.plot(t, sig)
# #             plt.plot(t, np.sum([1/x*np.sin(t*x*2*np.pi) for x in [1]],0))
# #             
#             plt.plot(t, sig_lp1)
#             plt.plot(t, sig_lp2)
#             #plt.plot(t, sig_lp3)
#               
#             plt.show()
#                
# #             plt.plot(tF, np.abs(F))
# #             plt.plot(tF, np.abs(F_lp1))
# # #             plt.plot(tF, np.abs(F_lp3))
# # #             plt.plot(tF, np.abs(F_lp2))
# #             plt.plot([0,1000],[0.708,.708])              
# #             plt.xlim([0,205])
# #             plt.ylim([0,1])
# #             plt.show()
# #             plt.plot(w1, 20 * np.log10(abs(h1)))
# #             plt.plot(w2, 20 * np.log10(abs(h2)))
# #             plt.xscale('log')
# #             plt.title('Butterworth filter frequency response')
# #             plt.xlabel('Frequency [radians / second]')
# #             plt.ylabel('Amplitude [dB]')
# #             plt.margins(0, 0.1)
# #             plt.grid(which='both', axis='both')
# #             plt.axvline(100, color='green') # cutoff frequency
# #             plt.show()
 
          
 
# #     def test_low_pass2(self):
# #         F = [0,1,0,0,0,0,0,0,0,0,0,0,0,0,.1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,.1]
# #         F +=[0]*(len(F)-1)
# #         x = F2x(F)
# #         a = np.tile(x, 3)
# #         print (x.shape)
# # #         a = np.random.randint(0,100,100).astype(np.float)
# #         b = first_order.low_pass(a, .1, 1)
# #         bb, ba = signal.butter(10,100, 'low', analog=True)
# #         #c = signal.lfilter(bb,ba, a)
# #         w, h = signal.freqs(bb, ba)
# #          
# # #         self.assertLess(b.std(), a.std())
# #         if 1:
# #             import matplotlib.pyplot as plt
# #             plt.plot(a)
# #             #plt.ylim([-2,2])
# #             plt.plot(b)
# #             #plt.plot(c)
# #             plt.show()
# #             print (h)
# #             plt.plot(w, abs(h))
# #             plt.xscale('log')
# #             plt.title('Butterworth filter frequency response')
# #             plt.xlabel('Frequency [radians / second]')
# #             plt.ylabel('Amplitude [dB]')
# #             plt.margins(0, 0.1)
# #             plt.grid(which='both', axis='both')
# #             plt.axvline(100, color='green') # cutoff frequency
# #             plt.show()
# #              
# #             
    

    def test_high_pass(self):
        a = np.random.randint(0,100,100).astype(np.float)
        b = first_order.high_pass(a, 1, 1)
        self.assertLess(b.mean(), a.mean())
        if 0:
            import matplotlib.pyplot as plt
            plt.plot(a)
            plt.plot(b)
            plt.show()
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 04/03/2013
@author: mmpe


'eq_load' calculate equivalent loads using one of the two rain flow counting methods
'cycle_matrix' calculates a matrix of cycles (binned on amplitude and mean value)
'eq_load_and_cycles' is used to calculate eq_loads of multiple time series (e.g. life time equivalent load)

The methods uses the rainflow counting routines (See documentation in top of methods):
- 'rainflow_windap': (Described in "Recommended Practices for Wind Turbine Testing - 3. Fatigue Loads",
                      2. edition 1990, Appendix A)
or
- 'rainflow_astm' (based on the c-implementation by Adam Nieslony found at the MATLAB Central File Exchange
                   http://www.mathworks.com/matlabcentral/fileexchange/3026)
'''
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from future import standard_library
import warnings
standard_library.install_aliases()
import numpy as np
from wetb.fatigue_tools.rainflowcounting import rainflowcount

rainflow_windap = rainflowcount.rainflow_windap
rainflow_astm = rainflowcount.rainflow_astm


def eq_load(signals, no_bins=46, m=[3, 4, 6, 8, 10, 12], neq=1, rainflow_func=rainflow_windap):
    """Equivalent load calculation

    Calculate the equivalent loads for a list of Wohler exponent and number of equivalent loads

    Parameters
    ----------
    signals : list of tuples or array_like
        - if list of tuples: list must have format [(sig1_weight, sig1),(sig2_weight, sig1),...] where\n
            - sigx_weight is the weight of signal x\n
            - sigx is signal x\n
        - if array_like: The signal
    no_bins : int, optional
        Number of bins in rainflow count histogram
    m : int, float or array-like, optional
        Wohler exponent (default is [3, 4, 6, 8, 10, 12])
    neq : int, float or array-like, optional
        The equivalent number of load cycles (default is 1, but normally the time duration in seconds is used)
    rainflow_func : {rainflow_windap, rainflow_astm}, optional
        The rainflow counting function to use (default is rainflow_windap)

    Returns
    -------
    eq_loads : array-like
        List of lists of equivalent loads for the corresponding equivalent number(s) and Wohler exponents

    Examples
    --------
    >>> signal = np.array([-2.0, 0.0, 1.0, 0.0, -3.0, 0.0, 5.0, 0.0, -1.0, 0.0, 3.0, 0.0, -4.0, 0.0, 4.0, 0.0, -2.0])
    >>> eq_load(signal, no_bins=50, neq=[1, 17], m=[3, 4, 6], rainflow_func=rainflow_windap)
    [[10.311095426959747, 9.5942535021382174, 9.0789213365013932], # neq = 1, m=[3,4,6]
    [4.010099657859783, 4.7249689509841746, 5.6618639965313005]], # neq = 17, m=[3,4,6]

    eq_load([(.4, signal), (.6, signal)], no_bins=50, neq=[1, 17], m=[3, 4, 6], rainflow_func=rainflow_windap)
    [[10.311095426959747, 9.5942535021382174, 9.0789213365013932], # neq = 1, m=[3,4,6]
    [4.010099657859783, 4.7249689509841746, 5.6618639965313005]], # neq = 17, m=[3,4,6]
    """
    try:
        return eq_load_and_cycles(signals, no_bins, m, neq, rainflow_func)[0]
    except TypeError:
        return [[np.nan] * len(np.atleast_1d(m))] * len(np.atleast_1d(neq))


def eq_load_and_cycles(signals, no_bins=46, m=[3, 4, 6, 8, 10, 12], neq=[10 ** 6, 10 ** 7, 10 ** 8], rainflow_func=rainflow_windap):
    """Calculate combined fatigue equivalent load

    Parameters
    ----------
    signals : list of tuples or array_like
        - if list of tuples: list must have format [(sig1_weight, sig1),(sig2_weight, sig1),...] where\n
            - sigx_weight is the weight of signal x\n
            - sigx is signal x\n
        - if array_like: The signal
    no_bins : int, optional
        Number of bins for rainflow counting
    m : int, float or array-like, optional
        Wohler exponent (default is [3, 4, 6, 8, 10, 12])
    neq : int or array-like, optional
        Equivalent number, default is [10^6, 10^7, 10^8]
    rainflow_func : {rainflow_windap, rainflow_astm}, optional
        The rainflow counting function to use (default is rainflow_windap)

    Returns
    -------
    eq_loads : array-like
        List of lists of equivalent loads for the corresponding equivalent number(s) and Wohler exponents
    cycles : array_like
        2d array with shape = (no_ampl_bins, 1)
    ampl_bin_mean : array_like
        mean amplitude of the bins
    ampl_bin_edges
        Edges of the amplitude bins
    """
    cycles, ampl_bin_mean, ampl_bin_edges, _, _ = cycle_matrix(signals, no_bins, 1, rainflow_func)
    if 0:  #to be similar to windap
        ampl_bin_mean = (ampl_bin_edges[:-1] + ampl_bin_edges[1:]) / 2
    cycles, ampl_bin_mean = cycles.flatten(), ampl_bin_mean.flatten()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        eq_loads = [[((np.nansum(cycles * ampl_bin_mean ** _m) / _neq) ** (1. / _m)) for _m in np.atleast_1d(m)]  for _neq in np.atleast_1d(neq)]
    return eq_loads, cycles, ampl_bin_mean, ampl_bin_edges


def cycle_matrix(signals, ampl_bins=10, mean_bins=10, rainflow_func=rainflow_windap):
    """Markow load cycle matrix

    Calculate the Markow load cycle matrix

    Parameters
    ----------
    Signals : array-like or list of tuples
        - if array-like, the raw signal\n
        - if list of tuples, list of (weight, signal), e.g. [(0.1,sig1), (0.8,sig2), (.1,sig3)]\n
    ampl_bins : int or array-like, optional
        if int, Number of amplitude value bins (default is 10)
        if array-like, the bin edges for amplitude
    mean_bins : int or array-like, optional
        if int, Number of mean value bins (default is 10)
        if array-like, the bin edges for mea
    rainflow_func : {rainflow_windap, rainflow_astm}, optional
        The rainflow counting function to use (default is rainflow_windap)

    Returns
    -------
    cycles : ndarray, shape(ampl_bins, mean_bins)
        A bi-dimensional histogram of load cycles(full cycles). Amplitudes are\
        histogrammed along the first dimension and mean values are histogrammed along the second dimension.
    ampl_bin_mean : ndarray, shape(ampl_bins,)
        The average cycle amplitude of the bins
    ampl_edges : ndarray, shape(ampl_bins+1,)
        The amplitude bin edges
    mean_bin_mean : ndarray, shape(ampl_bins,)
        The average cycle mean of the bins
    mean_edges : ndarray, shape(mean_bins+1,)
        The mean bin edges

    Examples
    --------
    >>> signal = np.array([-2.0, 0.0, 1.0, 0.0, -3.0, 0.0, 5.0, 0.0, -1.0, 0.0, 3.0, 0.0, -4.0, 0.0, 4.0, 0.0, -2.0])
    >>> cycles, ampl_bin_mean, ampl_edges, mean_bin_mean, mean_edges = cycle_matrix(signal)
    >>> cycles, ampl_bin_mean, ampl_edges, mean_bin_mean, mean_edges = cycle_matrix([(.4, signal), (.6,signal)])
    """

    if isinstance(signals[0], tuple):
        weights, ampls, means = np.array([(np.zeros_like(ampl)+weight,ampl,mean) for weight, signal in signals for ampl,mean in rainflow_func(signal[:]).T], dtype=np.float64).T
    else:
        ampls, means = rainflow_func(signals[:])
        weights = np.ones_like(ampls)
    if isinstance(ampl_bins, int):
        ampl_bins = np.linspace(0, 1, num=ampl_bins + 1) * ampls[weights>0].max()
    cycles, ampl_edges, mean_edges = np.histogram2d(ampls, means, [ampl_bins, mean_bins], weights=weights)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ampl_bin_sum = np.histogram2d(ampls, means, [ampl_bins, mean_bins], weights=weights * ampls)[0]
        ampl_bin_mean = np.nanmean(ampl_bin_sum / np.where(cycles,cycles,np.nan),1)
        mean_bin_sum = np.histogram2d(ampls, means, [ampl_bins, mean_bins], weights=weights * means)[0]
        mean_bin_mean = np.nanmean(mean_bin_sum / np.where(cycles, cycles, np.nan), 1)
    cycles = cycles / 2  # to get full cycles
    return cycles, ampl_bin_mean, ampl_edges, mean_bin_mean, mean_edges


def cycle_matrix2(signal, nrb_amp, nrb_mean, rainflow_func=rainflow_windap):
    """
    Same as wetb.fatigue_tools.fatigue.cycle_matrix but bin from min_amp to
    max_amp instead of 0 to max_amp.

    Parameters
    ----------

    Signal : ndarray(n)
        1D Raw signal array

    nrb_amp : int
        Number of bins for the amplitudes

    nrb_mean : int
        Number of bins for the means

    rainflow_func : {rainflow_windap, rainflow_astm}, optional
        The rainflow counting function to use (default is rainflow_windap)

    Returns
    -------

    cycles : ndarray, shape(ampl_bins, mean_bins)
        A bi-dimensional histogram of load cycles(full cycles). Amplitudes are\
        histogrammed along the first dimension and mean values are histogrammed
        along the second dimension.

    ampl_edges : ndarray, shape(no_bins+1,n)
        The amplitude bin edges

    mean_edges : ndarray, shape(no_bins+1,n)
        The mean bin edges

    """
    bins = [nrb_amp, nrb_mean]
    ampls, means = rainflow_func(signal)
    weights = np.ones_like(ampls)
    cycles, ampl_edges, mean_edges = np.histogram2d(ampls, means, bins,
                                                    weights=weights)
    cycles = cycles / 2  # to get full cycles

    return cycles, ampl_edges, mean_edges


if __name__ == "__main__":
    signal1 = np.array([-2.0, 0.0, 1.0, 0.0, -3.0, 0.0, 5.0, 0.0, -1.0, 0.0, 3.0, 0.0, -4.0, 0.0, 4.0, 0.0, -2.0])
    signal2 = signal1 * 1.1

    # equivalent load for default wohler slopes
    print (eq_load(signal1, no_bins=50, neq=17, rainflow_func=rainflow_windap))
    print (eq_load(signal1, no_bins=50, neq=17, rainflow_func=rainflow_astm))

    # Cycle matrix with 4 amplitude bins and 4 mean value bins
    print (cycle_matrix(signal1, 4, 4, rainflow_func=rainflow_windap))
    print (cycle_matrix(signal1, 4, 4, rainflow_func=rainflow_astm))

    # Cycle matrix where signal1 and signal2 contributes with 50% each
    print (cycle_matrix([(.5, signal1), (.5, signal2)], 4, 8, rainflow_func=rainflow_astm))




from ._io import load, read_sensor_info
from wetb import gtsdf
import numpy as np

class Dataset(gtsdf.Dataset):
    def __init__(self, filename, name_stop=8,dtype=np.float):
        self.time, self.data, self.info = load(filename, name_stop,dtype)


'''
Created on 04/12/2015

@author: mmpe
'''
import unittest
import matplotlib.pyplot as plt
import numpy as np
from wetb.wind import weibull

class TestWeibull(unittest.TestCase):


    def test_weibull(self):
        wb = weibull.pdf(4, 2)
        x = np.arange(0, 20, .1)
        self.assertAlmostEqual(wb(x).max(), 0.21441923796454562)
        if 0:
            plt.plot(x, wb(x))
            plt.show()

    def test_random_weibull(self):
        y = weibull.random(4, 2, 1000000)
        pdf, x = np.histogram(y, 100, normed=True)
        x = (x[1:] + x[:-1]) / 2
        self.assertLess(sum((pdf - weibull.pdf(4, 2)(x)) ** 2), 0.0001)
        if 0:
            plt.plot(x, pdf)
            plt.plot(x, weibull.pdf(4, 2)(x))
            plt.show()
 
    def test_fit_weibull(self):
        y = weibull.random(4, 2, 1000000)
        A, k = weibull.fit(y)
        self.assertAlmostEqual(A, 4, delta=0.01)
        self.assertAlmostEqual(k, 2, delta=0.01)
        if 0:
            plt.hist(y, 100, normed=True)
            x = np.arange(0, 20, .1)
            plt.plot(x, weibull.pdf(4, 2)(x))
            plt.show()
             
    def test_weibull_cdf(self):
        wb = weibull.pdf(4, 2)
        x = np.arange(0, 20, .01)
        cdf = weibull.cdf(4,2)
        self.assertEqual(cdf(999),1)
        np.testing.assert_array_almost_equal(np.cumsum(wb(x))/100, cdf(x), 3)
         
        if 0:
            plt.plot(x, np.cumsum(wb(x))*.01)
            plt.plot(x, cdf(x))
            plt.show()
             
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testweibull']
    unittest.main()



from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from io import open
from builtins import map
from builtins import range
from builtins import chr
from future import standard_library
standard_library.install_aliases()
import os
import numpy as np
import struct
from itertools import takewhile

def load_output(filename):
    """Load a FAST binary or ascii output file

    Parameters
    ----------
    filename : str
        filename

    Returns
    -------
    data : ndarray
        data values
    info : dict
        info containing:
            - name: filename
            - description: description of dataset
            - attribute_names: list of attribute names
            - attribute_units: list of attribute units
    """

    assert os.path.isfile(filename), "File, %s, does not exists" % filename
    with open(filename, 'r') as f:
        try:
            f.readline()
        except UnicodeDecodeError:
            return load_binary_output(filename)
    return load_ascii_output(filename)

def load_ascii_output(filename):
    with open(filename) as f:
        info = {}
        info['name'] = os.path.splitext(os.path.basename(filename))[0]
        # Header is whatever is before the keyword `time`
        in_header = True
        header = []
        while in_header:
            l = f.readline()
            if not l:
                raise Exception('Error finding the end of FAST out file header. Keyword Time missing.')
            in_header= (l+' dummy').lower().split()[0] != 'time'
            if in_header:
                header.append(l)
            else:
                info['description'] = header
                info['attribute_names'] = l.split()
                info['attribute_units'] = [unit[1:-1] for unit in f.readline().split()]

        # Data, up to end of file or empty line (potential comment line at the end)
        data = np.array([l.strip().split() for l in takewhile(lambda x: len(x.strip())>0, f.readlines())]).astype(np.float)
        return data, info


def load_binary_output(filename, use_buffer=True):
    """
    03/09/15: Ported from ReadFASTbinary.m by Mads M Pedersen, DTU Wind
    24/10/18: Low memory/buffered version by E. Branlard, NREL

    Info about ReadFASTbinary.m:
    % Author: Bonnie Jonkman, National Renewable Energy Laboratory
    % (c) 2012, National Renewable Energy Laboratory
    %
    %  Edited for FAST v7.02.00b-bjj  22-Oct-2012
    """
    def fread(fid, n, type):
        fmt, nbytes = {'uint8': ('B', 1), 'int16':('h', 2), 'int32':('i', 4), 'float32':('f', 4), 'float64':('d', 8)}[type]
        return struct.unpack(fmt * n, fid.read(nbytes * n))

    def freadRowOrderTableBuffered(fid, n, type_in, nCols, nOff=0, type_out='float64'):
        """ 
        Reads of row-ordered table from a binary file.

        Read `n` data of type `type_in`, assumed to be a row ordered table of `nCols` columns.
        Memory usage is optimized by allocating the data only once.
        Buffered reading is done for improved performances (in particular for 32bit python)

        `nOff` allows for additional column space at the begining of the storage table.
        Typically, `nOff=1`, provides a column at the beginning to store the time vector.

        @author E.Branlard, NREL

        """
        fmt, nbytes = {'uint8': ('B', 1), 'int16':('h', 2), 'int32':('i', 4), 'float32':('f', 4), 'float64':('d', 8)}[type_in]
        nLines          = int(n/nCols)
        GoodBufferSize  = 4096*40
        nLinesPerBuffer = int(GoodBufferSize/nCols)
        BufferSize      = nCols * nLinesPerBuffer
        nBuffer         = int(n/BufferSize)
        # Allocation of data
        data = np.zeros((nLines,nCols+nOff), dtype = type_out)
        # Reading
        try:
            nIntRead   = 0
            nLinesRead = 0
            while nIntRead<n:
                nIntToRead = min(n-nIntRead, BufferSize)
                nLinesToRead = int(nIntToRead/nCols)
                Buffer = np.array(struct.unpack(fmt * nIntToRead, fid.read(nbytes * nIntToRead)))
                Buffer = Buffer.reshape(-1,nCols)
                data[ nLinesRead:(nLinesRead+nLinesToRead),  nOff:(nOff+nCols)  ] = Buffer
                nLinesRead = nLinesRead + nLinesToRead
                nIntRead   = nIntRead   + nIntToRead
        except:
            raise Exception('Read only %d of %d values in file:' % (nIntRead, n, filename))
        return data


    FileFmtID_WithTime = 1  #% File identifiers used in FAST
    FileFmtID_WithoutTime = 2
    LenName = 10  #;  % number of characters per channel name
    LenUnit = 10  #;  % number of characters per unit name

    with open(filename, 'rb') as fid:
        FileID = fread(fid, 1, 'int16')  #;             % FAST output file format, INT(2)
        if FileID[0] not in [FileFmtID_WithTime, FileFmtID_WithoutTime]:
            raise Exception('FileID not supported {}. Is it a FAST binary file?'.format(FileID))

        NumOutChans = fread(fid, 1, 'int32')[0]  #;             % The number of output channels, INT(4)
        NT = fread(fid, 1, 'int32')[0]  #;             % The number of time steps, INT(4)


        if FileID == FileFmtID_WithTime:
            TimeScl = fread(fid, 1, 'float64')  #;           % The time slopes for scaling, REAL(8)
            TimeOff = fread(fid, 1, 'float64')  #;           % The time offsets for scaling, REAL(8)
        else:
            TimeOut1 = fread(fid, 1, 'float64')  #;           % The first time in the time series, REAL(8)
            TimeIncr = fread(fid, 1, 'float64')  #;           % The time increment, REAL(8)




        ColScl = fread(fid, NumOutChans, 'float32')  #; % The channel slopes for scaling, REAL(4)
        ColOff = fread(fid, NumOutChans, 'float32')  #; % The channel offsets for scaling, REAL(4)

        LenDesc = fread(fid, 1, 'int32')[0]  #;  % The number of characters in the description string, INT(4)
        DescStrASCII = fread(fid, LenDesc, 'uint8')  #;  % DescStr converted to ASCII
        DescStr = "".join(map(chr, DescStrASCII)).strip()



        ChanName = []  # initialize the ChanName cell array
        for iChan in range(NumOutChans + 1):
            ChanNameASCII = fread(fid, LenName, 'uint8')  #; % ChanName converted to numeric ASCII
            ChanName.append("".join(map(chr, ChanNameASCII)).strip())


        ChanUnit = []  # initialize the ChanUnit cell array
        for iChan in range(NumOutChans + 1):
            ChanUnitASCII = fread(fid, LenUnit, 'uint8')  #; % ChanUnit converted to numeric ASCII
            ChanUnit.append("".join(map(chr, ChanUnitASCII)).strip()[1:-1])


        #    %-------------------------
        #    % get the channel time series
        #    %-------------------------

        nPts = NT * NumOutChans  #;           % number of data points in the file


        if FileID == FileFmtID_WithTime:
            PackedTime = fread(fid, NT, 'int32')  #; % read the time data
            cnt = len(PackedTime)
            if cnt < NT:
                raise Exception('Could not read entire %s file: read %d of %d time values' % (filename, cnt, NT))

        if use_buffer:
            # Reading data using buffers, and allowing an offset for time column (nOff=1)
            data = freadRowOrderTableBuffered(fid, nPts, 'int16', NumOutChans, nOff=1, type_out='float64')
        else:
            # NOTE: unpacking huge data not possible on 32bit machines
            PackedData = fread(fid, nPts, 'int16')  #; % read the channel data
            cnt = len(PackedData)
            if cnt < nPts:
                raise Exception('Could not read entire %s file: read %d of %d values' % (filename, cnt, nPts))
            data = np.array(PackedData).reshape(NT, NumOutChans)
            del PackedData

    if FileID == FileFmtID_WithTime:
        time = (np.array(PackedTime) - TimeOff) / TimeScl;
    else:
        time = TimeOut1 + TimeIncr * np.arange(NT)

    #    %-------------------------
    #    % Scale the packed binary to real data
    #    %-------------------------
    if use_buffer:
        # Scaling Data
        for iCol in range(NumOutChans):
            data[:,iCol+1] = (data[:,iCol+1] - ColOff[iCol]) / ColScl[iCol]
        # Adding time column
        data[:,0] = time
    else:
        # NOTE: memory expensive due to time conversion, and concatenation
        data = (data - ColOff) / ColScl
        data = np.concatenate([time.reshape(NT, 1), data], 1)

    info = {'name': os.path.splitext(os.path.basename(filename))[0],
            'description': DescStr,
            'attribute_names': ChanName,
            'attribute_units': ChanUnit}
    return data, info




# -*- coding: utf-8 -*-
"""
Created on Thu Aug 04 09:24:51 2016

@author: tlbl
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import numpy as np


class Control(object):

    def torque_controller_tuning(self, r0, lambda_opt, ksi_pole_part,
                                 omega_pole_part, j_dt):
        """THIS IS A WIP
        """

        # TODO: THIS IS STILL A WIP

        i12, i23, i34 = self.select_regions()

        # Compute K factor for optimal CP tracking in region 1 as
        # K=0.5*rho*A*Cp_opt*R^3/lambda_opt^3

        # HAWCStab2/Computation_LIB/compute_controller_input_mod.f90:83-88

        # Compute K factor for optimal CP tracking in region 1 as
        # K=0.5*rho*A*Cp_opt*R^3/lambda_opt^3
        # pi_torque_controller%kfactor=0.d0
        # do i=1,iregion2-iregion1-1
        # !kfactor=kfactor+0.85d0*kfactor_fit(i)*substructure(3)
        #   pi_torque_controller%kfactor
        #   = pi_torque_controller%kfactor
        #   + kfactor_fit(i)*substructure(3)%opstate(i+iregion1)%r0**3/lambda_opt**3
        # enddo

        # HAWCStab2/Computation_LIB/compute_controller_input_mod.f90:89
        # pi_torque_controller%kfactor =
        #   pi_torque_controller%kfactor/max(dfloat(iregion2-iregion1-1), 1.d0)

        Qg = (r0*r0*r0)/(lambda_opt*lambda_opt*lambda_opt)

        # PI generator torque controller in region 2 (constant speed, variable power)
        # HAWCStab2/Computation_LIB/compute_controller_input_mod.f90:104-105
        # pi_torque_controller%kp = 2.d0*ksi_pole_part*omega_pole_part*j_dt
        # pi_torque_controller%ki = omega_pole_part**2*j_dt
        pgTorque = 2.0*ksi_pole_part*omega_pole_part*j_dt
        igTorque = omega_pole_part**2*j_dt

        return Qg, pgTorque, igTorque


    def pitch_controller_tuning(self, pitch, I, dQdt, P, Omr, om, csi):
        """

        Function to compute the gains of the pitch controller of the Basic DTU
        Wind Energy Controller with the pole placement technique implemented
        in HAWCStab2.

        Parameters
        ----------
        pitch: array
            Pitch angle [deg]. The values should only be those of the full load
            region.
        I: array
            Drivetrain inertia [kg*m**2]
        dQdt: array
            Partial derivative of the aerodynamic torque with respect to the
            pitch angle [kNm/deg]. Can be computed with HAWCStab2.
        P: float
            Rated power [kW]. Set to zero in case of constant torque regulation
        Omr: float
            Rated rotational speed [rpm]
        om: float
            Freqeuncy of regulator mode [Hz]
        csi: float
            Damping ratio of regulator mode

        Returns
        -------
        kp: float
            Proportional gain [rad/(rad/s)]
        ki: float
            Intagral gain [rad/rad]
        K1: float
            Linear term of the gain scheduling [deg]
        K2: float
            Quadratic term of the gain shceduling [deg**2]


        """
        pitch = pitch * np.pi/180.
        I = I * 1e-3
        dQdt = dQdt * 180./np.pi
        Omr = Omr * np.pi/30.
        om = om * 2.*np.pi

        # Quadratic fitting of dQdt
        A = np.ones([dQdt.shape[0], 3])
        A[:, 0] = pitch**2
        A[:, 1] = pitch
        b = dQdt
        ATA = np.dot(A.T, A)
        iATA = np.linalg.inv(ATA)
        iATAA = np.dot(iATA, A.T)
        x = np.dot(iATAA, b)

        kp = -(2*csi*om*I[0] - P/(Omr**2))/x[2]
        ki = -(om**2*I[0])/x[2]

        K1 = x[2]/x[1]*(180./np.pi)
        K2 = x[2]/x[0]*(180./np.pi)**2

        return kp, ki, K1, K2

    def K_omega2(V, P, R, TSR):

        Va = np.array(V)
        Pa = np.array(P)
        Ra = np.array(R)
        TSRa = np.array(TSR)
        K = Ra**3 * np.mean(Pa/(TSRa*Va)**3)

        return K

    def select_regions(self, pitch, omega, power):
        """Find indices at wich point the controller should switch between the
        different operating mode regions.

        Parameters
        ----------



        Returns
        ------

        i12 : int

        123 : int

        134 : int

        """

        i12 = 0

        n = len(pitch)

        for i in range(n-1):
            if (abs(power[i]/power[i+1] - 1.) > 0.01):
                if (abs(omega[i] / omega[i+1] - 1.) > 0.01):
                    i12 = i
                    break
        i23 = n-1
        for i in range(i12, n-1):
            if (abs(omega[i] / omega[i+1] - 1.) < 0.01):
                i23 = i
                break

        i34 = i23
        for i in range(i23, n-1):
            if (abs(power[i]/power[i+1] - 1.) > 0.01):
                if (abs(omega[i] / omega[i+1] - 1.) < 0.01):
                    i34 = i+1

        return i12, i23, i34


if __name__ == '__main__':

    pass



'''
Created on 05/11/2015

@author: MMPE
'''
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from io import open
from future import standard_library
from wetb.hawc2.shear_file import ShearFile
standard_library.install_aliases()
import unittest
from wetb.hawc2 import shear_file
import numpy as np
import os
import shutil
tfp = os.path.join(os.path.dirname(__file__), 'test_files/')
class TestShearFile(unittest.TestCase):


    def test_shearfile_save(self):
        f = tfp + "tmp_shearfile1.dat"
        shear_file.save(f, [-55, 55], [30, 100, 160] , u=np.array([[0.7, 1, 1.3], [0.7, 1, 1.3]]).T)
        with open(f) as fid:
            self.assertEqual(fid.read(),
""" # autogenerated shear file
  2 3
 # shear v component
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
 # shear u component
  0.7000000000 0.7000000000
  1.0000000000 1.0000000000
  1.3000000000 1.3000000000
 # shear w component
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
 # v coordinates
  -55.0000000000
  55.0000000000
 # w coordinates
  30.0000000000
  100.0000000000
  160.0000000000
""")
        os.remove(f)


    def test_shearfile2(self):
        f = tfp + "tmp_shearfile2.dat"
        shear_file.save(f, [-55, 55], [30, 100, 160] , u=np.array([0.7, 1, 1.3]).T)
        with open(f) as fid:
            self.assertEqual(fid.read(),
""" # autogenerated shear file
  2 3
 # shear v component
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
 # shear u component
  0.7000000000 0.7000000000
  1.0000000000 1.0000000000
  1.3000000000 1.3000000000
 # shear w component
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
  0.0000000000 0.0000000000
 # v coordinates
  -55.0000000000
  55.0000000000
 # w coordinates
  30.0000000000
  100.0000000000
  160.0000000000
""")
        os.remove(f)

    def test_shear_makedirs(self):
        f = tfp + "shear/tmp_shearfile2.dat"
        shear_file.save(f, [-55, 55], [30, 100, 160] , u=np.array([0.7, 1, 1.3]).T)
        shutil.rmtree(tfp + "shear")
        
    def test_shear_load(self):
        shear_file = ShearFile.load(tfp+"data/user_shear.dat")
        np.testing.assert_array_equal(shear_file.w_positions, [30,100,160])
        self.assertEqual(shear_file.uvw(0,65)[0],.85)
        self.assertEqual(shear_file.uvw(-55,65)[0],.9)
        np.testing.assert_array_equal(shear_file.uvw([0,-55],[65,65])[0],[.85,.9])
        
        shear_file = ShearFile.load_from_htc(tfp+"htcfiles/test.htc")
        np.testing.assert_array_equal(shear_file.w_positions, [30,100,160])
        np.testing.assert_array_almost_equal(shear_file.uvw([0,-55],[65,65])[0],np.array([.85,.9])*10+8.860807038)
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.test_shearfile']
    unittest.main()



'''
Created on 27/11/2015

@author: MMPE
'''

from io import StringIO
import sys
import paramiko

import os
import threading
from _collections import deque
import time
import traceback
import zipfile
import glob
from sshtunnel import SSHTunnelForwarder, SSH_CONFIG_FILE
from wetb.utils.ui import UI


class SSHInteractiveAuthTunnelForwarder(SSHTunnelForwarder):
    def __init__(
            self,
            interactive_auth_handler,
            ssh_address_or_host=None,
            ssh_config_file=SSH_CONFIG_FILE,
            ssh_host_key=None,
            ssh_password=None,
            ssh_pkey=None,
            ssh_private_key_password=None,
            ssh_proxy=None,
            ssh_proxy_enabled=True,
            ssh_username=None,
            local_bind_address=None,
            local_bind_addresses=None,
            logger=None,
            mute_exceptions=False,
            remote_bind_address=None,
            remote_bind_addresses=None,
            set_keepalive=0.0,
            threaded=True,
            compression=None,
            allow_agent=True, *
            args, **
            kwargs):
        self.interactive_auth_handler = interactive_auth_handler
        SSHTunnelForwarder.__init__(self, ssh_address_or_host=ssh_address_or_host, ssh_config_file=ssh_config_file, ssh_host_key=ssh_host_key, ssh_password=ssh_password, ssh_pkey=ssh_pkey, ssh_private_key_password=ssh_private_key_password, ssh_proxy=ssh_proxy, ssh_proxy_enabled=ssh_proxy_enabled, ssh_username=ssh_username,
                                    local_bind_address=local_bind_address, local_bind_addresses=local_bind_addresses, logger=logger, mute_exceptions=mute_exceptions, remote_bind_address=remote_bind_address, remote_bind_addresses=remote_bind_addresses, set_keepalive=set_keepalive, threaded=threaded, compression=compression, allow_agent=allow_agent, *args, **kwargs)

    def _connect_to_gateway(self):
        """
        Open connection to SSH gateway
         - First try with all keys loaded from an SSH agent (if allowed)
         - Then with those passed directly or read from ~/.ssh/config
         - As last resort, try with a provided password
        """
        try:
            self._transport = self._get_transport()
            self._transport.start_client()
            self._transport.auth_interactive(self.ssh_username, self.interactive_auth_handler)
            if self._transport.is_alive:
                return
        except paramiko.AuthenticationException:
            self.logger.debug('Authentication error')
            self._stop_transport()

        self.logger.error('Could not open connection to gateway')

    def _connect_to_gateway_old(self):
        """
        Open connection to SSH gateway
         - First try with all keys loaded from an SSH agent (if allowed)
         - Then with those passed directly or read from ~/.ssh/config
         - As last resort, try with a provided password
        """
        if self.ssh_password:  # avoid conflict using both pass and pkey
            self.logger.debug('Trying to log in with password: {0}'
                              .format('*' * len(self.ssh_password)))
            try:
                self._transport = self._get_transport()
                if self.interactive_auth_handler:
                    self._transport.start_client()
                    self._transport.auth_interactive(self.ssh_username, self.interactive_auth_handler)
                else:
                    self._transport.connect(hostkey=self.ssh_host_key,
                                            username=self.ssh_username,
                                            password=self.ssh_password)

                if self._transport.is_alive:
                    return
            except paramiko.AuthenticationException:
                self.logger.debug('Authentication error')
                self._stop_transport()

        self.logger.error('Could not open connection to gateway')


class SSHClient(object):
    "A wrapper of paramiko.SSHClient"
    TIMEOUT = 4

    def __init__(self, host, username, password=None, port=22, key=None, passphrase=None, interactive_auth_handler=None, gateway=None, ui=UI()):
        self.host = host
        self.username = username
        self.password = password
        self.port = port
        self.key = key
        self.gateway = gateway
        self.interactive_auth_handler = interactive_auth_handler
        self.ui = ui
        self.disconnect = 0
        self.client = None
        self.ssh_lock = threading.RLock()
        self._sftp = None
        self.transport = None
        self.counter_lock = threading.RLock()
        self.counter = 0
        if key is not None:
            self.key = paramiko.RSAKey.from_private_key(StringIO(key), password=passphrase)

    def info(self):
        return self.host, self.username, self.password, self.port

    def __enter__(self):
        with self.ssh_lock:
            self.disconnect += 1
            if self.client is None or self.client._transport is None or self.client._transport.is_active() is False:
                self.close()
                try:
                    self.connect()
                    self.disconnect = 1
                except Exception as e:
                    self.close()
                    self.disconnect = 0
                    raise e
            return self.client

    def connect(self):
        if self.gateway:
            if self.gateway.interactive_auth_handler:
                self.tunnel = SSHInteractiveAuthTunnelForwarder(self.gateway.interactive_auth_handler,
                                                                (self.gateway.host, self.gateway.port),
                                                                ssh_username=self.gateway.username,
                                                                ssh_password=self.gateway.password,
                                                                remote_bind_address=(self.host, self.port),
                                                                local_bind_address=('0.0.0.0', 10022)
                                                                )
            else:
                self.tunnel = SSHTunnelForwarder((self.gateway.host, self.gateway.port),
                                                 ssh_username=self.gateway.username,
                                                 ssh_password=self.gateway.password,
                                                 remote_bind_address=(self.host, self.port),
                                                 local_bind_address=('0.0.0.0', 10022)
                                                 )

            print("start tunnel")
            self.tunnel.start()
            print("self.client = paramiko.SSHClient()")
            self.client = paramiko.SSHClient()

            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            print('self.client.connect("127.0.0.1", 10022, username=self.username, password=self.password)')
            self.client.connect("127.0.0.1", 10022, username=self.username, password=self.password)
            print("done")

        elif self.password is None or self.password == "":
            raise IOError("Password not set for %s" % self.host)
        else:
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            try:
                self.client.connect(self.host, self.port, username=self.username,
                                    password=self.password, pkey=self.key, timeout=self.TIMEOUT)
            except paramiko.ssh_exception.SSHException as e:
                transport = self.client.get_transport()
                transport.auth_interactive(self.username, self.interactive_auth_handler)

        assert self.client is not None
        #self.sftp = paramiko.SFTPClient.from_transport(self.client._transport)
        return self

    def __del__(self):
        self.close()

    @property
    def sftp(self):
        if self._sftp is None:
            self._sftp = paramiko.SFTPClient.from_transport(self.client._transport)
        return self._sftp


#     @sftp.setter
#     def sftp(self, values):
#         pass

    def __exit__(self, *args):
        self.disconnect -= 1
        if self.disconnect == 0:
            self.close()

    def download(self, remotefilepath, localfile, verbose=False, retry=1, callback=None):
        if verbose:
            ret = None
            print("Download %s > %s" % (remotefilepath, str(localfile)))
        if callback is None:
            callback = self.ui.progress_callback()

        for i in range(retry):
            if i > 0:
                print("Retry download %s, #%d" % (remotefilepath, i))

            try:
                SSHClient.__enter__(self)
                if isinstance(localfile, (str, bytes, int)):
                    ret = self.sftp.get(remotefilepath, localfile, callback=callback)
                elif hasattr(localfile, 'write'):
                    ret = self.sftp.putfo(remotefilepath, localfile, callback=callback)
                break
            except:
                pass
            finally:
                SSHClient.__exit__(self)

            print("Download %s failed from %s" % (remotefilepath, self.host))
        if verbose:
            print(ret)

    def upload(self, localfile, filepath, verbose=False, callback=None):
        if verbose:
            print("Upload %s > %s" % (localfile, filepath))
        if callback is None:
            callback = self.ui.progress_callback()
        try:
            SSHClient.__enter__(self)
            self.execute("mkdir -p %s" % (os.path.dirname(filepath)))
            sftp = self.sftp
            if isinstance(localfile, (str, bytes, int)):
                ret = sftp.put(localfile, filepath, callback=callback)
            elif hasattr(localfile, 'read'):
                size = len(localfile.read())
                localfile.seek(0)
                ret = sftp.putfo(localfile, filepath, file_size=size, callback=callback)
        except Exception as e:
            print("upload failed ", str(e))
            raise e
        finally:
            sftp.close()
            SSHClient.__exit__(self)
        if verbose:
            print(ret)

    def upload_files(self, localpath, remotepath, file_lst=["."], compression_level=1, callback=None):
        assert os.path.isdir(localpath)
        if not isinstance(file_lst, (tuple, list)):
            file_lst = [file_lst]
        files = ([os.path.join(root, f) for fp in file_lst for root, _, files in os.walk(os.path.join(localpath, fp)) for f in files] +
                 [f for fp in file_lst for f in glob.glob(os.path.join(localpath, fp))])
        files = set([os.path.abspath(f) for f in files])

        compression_levels = {0: zipfile.ZIP_STORED, 1: zipfile.ZIP_DEFLATED, 2: zipfile.ZIP_BZIP2, 3: zipfile.ZIP_LZMA}
        with self.counter_lock:
            self.counter += 1
            zn = 'tmp_%s_%04d.zip' % (id(self), self.counter)
        zipf = zipfile.ZipFile(zn, 'w', compression_levels[compression_level])
        try:
            for f in files:
                zipf.write(f, os.path.relpath(f, localpath))
            zipf.close()
            remote_zn = os.path.join(remotepath, zn).replace("\\", "/")
            with self:
                self.execute("mkdir -p %s" % (remotepath))

                self.upload(zn, remote_zn, callback=callback)
                self.execute("unzip %s -d %s && rm %s" % (remote_zn, remotepath, remote_zn))
        except:
            print("upload files failed", )
            traceback.print_exc()
            raise
        finally:
            os.remove(zn)

    def download_files(self, remote_path, localpath, file_lst=["."], compression_level=1, callback=None):
        if not isinstance(file_lst, (tuple, list)):
            file_lst = [file_lst]
        file_lst = [f.replace("\\", "/") for f in file_lst]
        with self.counter_lock:
            self.counter += 1
            zn = 'tmp_%s_%04d.zip' % (id(self), self.counter)

        remote_zip = os.path.join(remote_path, zn).replace("\\", "/")
        self.execute("cd %s && zip -r %s %s" % (remote_path, zn, " ".join(file_lst)))

        local_zip = os.path.join(localpath, zn)
        if not os.path.isdir(localpath):
            os.makedirs(localpath)
        self.download(remote_zip, local_zip, callback=callback)
        self.execute("rm -f %s" % remote_zip)
        with zipfile.ZipFile(local_zip, "r") as z:
            z.extractall(localpath)
        os.remove(local_zip)

    def close(self):
        for x in ["_sftp", "client",  "tunnel"]:
            try:
                getattr(self, x).close()
                setattr(self, x, None)
            except:
                pass
        self.disconnect = False

    def file_exists(self, filename):
        _, out, _ = (self.execute(
            '[ -f %s ] && echo "File exists" || echo "File does not exists"' % filename.replace("\\", "/")))
        return out.strip() == "File exists"

    def execute(self, command, sudo=False, verbose=False):
        feed_password = False
        if sudo and self.username != "root":
            command = "sudo -S -p '' %s" % command
            feed_password = self.password is not None and len(self.password) > 0
        if isinstance(command, (list, tuple)):
            command = "\n".join(command)

        if verbose:
            print(">>> " + command)
        with self as ssh:
            if ssh is None:
                exc_info = sys.exc_info()
                traceback.print_exception(*exc_info)
                raise Exception("ssh_client exe ssh is None")
            stdin, stdout, stderr = ssh.exec_command(command)
            if feed_password:
                stdin.write(self.password + "\n")
                stdin.flush()

            v, out, err = stdout.channel.recv_exit_status(), stdout.read().decode(), stderr.read().decode()

        if v:
            raise Warning("out:\n%s\n----------\nerr:\n%s" % (out, err))
        elif verbose:
            if out:
                sys.stdout.write(out)
            if err:
                sys.stderr.write(err)
        return v, out, err

    def append_wine_path(self, path):
        ret = self.execute(
            r'wine regedit /E tmp.reg "HKEY_LOCAL_MACHINE\System\CurrentControlSet\Control\Session Manager\Environment"')
        self.download('tmp.reg', 'tmp.reg')
        with open('tmp.reg') as fid:
            lines = fid.readlines()

        path_line = [l for l in lines if l.startswith('"PATH"=')][0]
        for p in path_line[8:-1].split(";"):
            if os.path.abspath(p) == os.path.abspath(p):
                return
        if path not in path_line:
            path_line = path_line.strip()[:-1] + ";" + path + '"'

            with open('tmp.reg', 'w') as fid:
                fid.write("".join(lines[:3] + [path_line]))
            self.upload('tmp.reg', 'tmp.reg')
            ret = self.execute('wine regedit tmp.reg')

    def glob(self, filepattern, cwd="", recursive=False):
        if isinstance(filepattern, list):
            with self:
                return [f for fp in filepattern for f in self.glob(fp, cwd, recursive)]
        cwd = os.path.join(cwd, os.path.split(filepattern)[0]).replace("\\", "/")
        filepattern = os.path.split(filepattern)[1]
        if recursive:
            _, out, _ = self.execute(r'find %s -type f -name "%s"' % (cwd, filepattern))
        else:
            _, out, _ = self.execute(r'find %s -maxdepth 1 -type f -name "%s"' % (cwd, filepattern))
        return [file for file in out.strip().split("\n") if file != ""]


class SharedSSHClient(SSHClient):
    def __init__(self, host, username, password=None, port=22, key=None, passphrase=None, interactive_auth_handler=None, gateway=None):
        SSHClient.__init__(self, host, username, password=password, port=port, key=key,
                           passphrase=passphrase, interactive_auth_handler=interactive_auth_handler, gateway=gateway)

        self.shared_ssh_queue = deque()
        self.next = None

    def execute(self, command, sudo=False, verbose=False):
        res = SSHClient.execute(self, command, sudo=sudo, verbose=verbose)
        return res

    def __enter__(self):
        with self.ssh_lock:
            SSHClient.__enter__(self)
            #print ("request SSH", threading.currentThread())
#             if len(self.shared_ssh_queue)>0 and self.shared_ssh_queue[0] == threading.get_ident():
#                 # SSH already allocated to this thread ( multiple use-statements in "with ssh:" block
#                 self.shared_ssh_queue.appendleft(threading.get_ident())
#             else:
#                 self.shared_ssh_queue.append(threading.get_ident())

            if len(self.shared_ssh_queue) > 0 and self.shared_ssh_queue[0] == threading.get_ident():
                # SSH already allocated to this thread ( multiple use-statements in "with ssh:" block
                self.shared_ssh_queue.popleft()

            self.shared_ssh_queue.append(threading.get_ident())

        while self.shared_ssh_queue[0] != threading.get_ident():
            time.sleep(2)

        return self.client

    def __exit__(self, *args):
        with self.ssh_lock:
            if len(self.shared_ssh_queue) > 0 and self.shared_ssh_queue[0] == threading.get_ident():
                self.shared_ssh_queue.popleft()


if __name__ == "__main__":
    try:
        import x
        username, password = 'mmpe', x.password
        client = SSHClient(host='jess.dtu.dk', port=22, username=username, password=password)
        print(client.execute("echo hello $USER from $HOSTNAME")[1])
    except ImportError:
        x = None



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
test = "TEST"
__version__ = '0.0.11'
# try:
#     import pkg_resources
#     __version__ = pkg_resources.safe_version(pkg_resources.get_distribution(__name__).version)
# except:
#     __version__ = 'unknown'





class OutputUI(object):
    def __init__(self, parent=None):
        self.parent = parent

    def run(self, f, *args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Warning as e:
            self.show_warning(e)
        except Exception as e:
            self.show_error(e)
            raise

    def show_error(self, msg, title="Error"):
        pass

    def show_message(self, msg, title="Information"):
        pass

    def show_warning(self, msg, title="Warning"):
        pass

    def show_text(self, text, end="\n", flush=False):
        pass

class InputUI(object):
    def get_confirmation(self, title, msg):
        raise NotImplementedError

    def get_string(self, title, msg):
        raise NotImplementedError

    def get_open_filename(self, title="Open", filetype_filter="*.*", file_dir=None, selected_filter=None):
        raise NotImplementedError

    def get_save_filename(self, title, filetype_filter, file_dir=None, selected_filter=None):
        raise NotImplementedError

    def get_open_filenames(self, title, filetype_filter, file_dir=None):
        raise NotImplementedError

    def get_foldername(self, title='Select directory', file_dir=None):
        raise NotImplementedError

class StatusUI(object):
    is_waiting = False
    def progress_iterator(self, sequence, text="Working... Please wait", allow_cancel=True):
        return sequence

    def exec_long_task(self, text, allow_cancel, task, *args, **kwargs):
        return task(*args, **kwargs)

    def progress_callback(self):
        def callback(n,N):
            pass 
        return callback

    def start_wait(self):
        self.is_waiting = True

    def end_wait(self):
        self.is_waiting = False

class UI(InputUI, OutputUI, StatusUI):
    pass



import glob
import os
import copy
from wetb.hawc2.hawc2_pbs_file import JESS_WINE32_HAWC2MB
from wetb.hawc2.htc_file import HTCFile
from wetb.utils.cluster_tools.pbsfile import PBSMultiRunner


class HTCFileSet():
    def __init__(self, model_path, htc_lst="**/*.htc"):
        self.model_path = model_path

        if not isinstance(htc_lst, list):
            htc_lst = [htc_lst]

        self.htc_files = []
        for htc_path in htc_lst:
            if os.path.isfile(htc_path):
                self.htc_files.append(htc_path)
            else:
                if not os.path.isabs(htc_path):
                    htc_path = os.path.join(model_path, htc_path)
                for filename in glob.iglob(htc_path, recursive=True):
                    self.htc_files.append(filename)

    def pbs_files(self, hawc2_path, hawc2_cmd, queue='workq', walltime=None,
                  input_files=None, output_files=None, copy_turb=(True, True)):

        return (HTCFile(htc).pbs_file(hawc2_path, hawc2_cmd, queue=queue, walltime=walltime,
                                      input_files=copy.copy(input_files),
                                      output_files=copy.copy(output_files),
                                      copy_turb=copy_turb) for htc in self.htc_files)

    def save_pbs_files(self, hawc2_path=None, hawc2_cmd=JESS_WINE32_HAWC2MB, queue='workq', walltime=None,
                       input_files=None, output_files=None, copy_turb=(True, True)):
        for pbs in self.pbs_files(hawc2_path, hawc2_cmd, queue=queue, walltime=walltime,
                                  input_files=input_files, output_files=output_files,
                                  copy_turb=copy_turb):
            pbs.save(self.model_path)


if __name__ == '__main__':
    #model_path = r'R:\HAWC2_tests\v12.6_mmpe3\win32\simple1'
    model_path = "w:/simple1"
    pbs_files = HTCFileSet(model_path).pbs_files(
        hawc2_path=r"R:\HAWC2_tests\v12.6_mmpe3\hawc2\win32", hawc2_cmd=JESS_WINE32_HAWC2MB, input_files=['data/*'])
    import pandas as pd
    time_overview = pd.read_excel(
        r'C:\mmpe\programming\Fortran\HAWC2_git\HAWC2\pytest_hawc2\release_tests\Time_overview.xlsx')
    for pbs in pbs_files:
        f = pbs.filename

        pbs.walltime = time_overview.loc[f[:-3].replace("pbs_in/", 'simple1/')]['mean'] * 24 * 3600
        pbs.save(model_path)
    PBSMultiRunner(model_path, nodes=1, ppn=10).save()



'''
Created on 13/10/2014

@author: MMPE
'''
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import zip
from future import standard_library
standard_library.install_aliases()

def bearing_damage(angle_moment_lst, m=3, thresshold=0.1):
    """Function ported from Matlab.

    Parameters
    ----------
    angle_moment_lst : ((angle_A_vector, moment_A_vector),(angle_B_vector, moment_B_vector),...)
        Angles[deg] and momements (e.g. pitch angle and blade root moment for the blades of a wind turbine)
    m : int, optional
        analogue to Wohler exponent, should be 3
    threeshold : float, optional
        Pitch noise. Pitch movement below this thresshold is ignored

    Returns
    -------
    max_damage : float
        A damage value of the most damaged pitch bearing. Only suitable for comparison
    """

    damages = []
    for angle, moment in angle_moment_lst:
        angle, moment = angle.tolist(), moment.tolist()
        k1 = 0
        p1, mflap = angle[0], moment[0]
        damage = 0
        for k, (pi, mo) in enumerate(zip(angle[1:], moment[1:]), 1):
            dangle = abs(pi - p1)
            if dangle > thresshold:
                damage += dangle * abs(mflap / (k - k1 + 1)) ** m
                k1 = k
                mflap = mo
                p1 = pi
            else:
                mflap += mo
        damages.append(damage)
    return max(damages)





import os


def repl(path):
    return path.replace("\\", "/")


def abspath(path):
    return repl(os.path.abspath(path))


def relpath(path, start=None):
    return repl(os.path.relpath(path, start))


def pjoin(*path):
    return repl(os.path.join(*path))


def cluster_path(path):
    try:
        import win32wnet
        drive, folder = os.path.splitdrive(abspath(path))
        path = abspath(win32wnet.WNetGetUniversalName(drive, 1) + folder)
    except Exception:
        path = repl(path)
    path = path.replace("//jess.dtu.dk", "/home")
    path = path.replace("//mimer.risoe.dk/aiolos", "/mnt/aiolos")
    path = path.replace("//mimer.risoe.dk", "/mnt/mimer")
    return path



'''
Created on 07/07/2015

@author: MMPE
'''
import numpy as np
from wetb.signal.fit import bin_fit


def fourier_fit(y, max_nfft, x=None):
    """Approximate a signal, y, with Fourier fit"""
    d = np.arange(360)
    return d, lambda deg : np.interp(deg%360, d, F2x(x2F(y, max_nfft, x)))

def fourier_fit_old(y, nfft):
    F = np.zeros(len(y), dtype=np.complex)
    F[:nfft + 1] = x2F(y, nfft)[:nfft + 1]
    return np.fft.ifft(F) * len(F)

def F2x(F_coefficients):
    """Compute signal from Fourier coefficients"""
    F = np.zeros(360, dtype=np.complex)
    nfft = len(F_coefficients) // 2
    F[:nfft + 1] = np.conj(F_coefficients[:nfft + 1])
    F[1:nfft + 1] += (F_coefficients[-nfft:][::-1])
    return np.real(np.fft.ifft(F) * len(F))

def x2F(y, max_nfft, x=None):
    """Compute Fourier coefficients from signal (signal may contain NANs)"""
    d = np.arange(360)
    if x is not None:
        x,fit = bin_fit(x,y, d)
        y = fit(d)
        
    nfft = min(max_nfft, len(y) // 2 + 1)
    n = len(y)
    N = nfft * 2 + 1
    theta = np.linspace(0, 2 * np.pi, n + 1)[:n]
    theta[np.isnan(y)] = np.nan
    a = np.empty((nfft * 2 + 1, nfft * 2 + 1))
    b = np.empty(nfft * 2 + 1)
    A0_lst = lambda dF : 2 * np.nansum(1 * dF)
    A_lst = lambda dF : [2 * np.nansum(np.cos(i * theta) * dF) for i in range(1, nfft + 1)]
    B_lst = lambda dF : [2 * np.nansum(np.sin(i * theta) * dF) for i in range(1, nfft + 1)]
    row = lambda dF : np.r_[A0_lst(dF), A_lst(dF), B_lst(dF)]


    for i in range(nfft + 1):
        a[i, :] = row(np.cos(i * theta))
        b[i] = 2 * np.nansum(y * np.cos(i * theta))
    for i, r in enumerate(range(nfft + 1, nfft * 2 + 1), 1):
        a[r, :] = row(np.sin(i * theta))
        b[r] = 2 * np.nansum(y * np.sin(i * theta))
    AB = np.linalg.solve(a, b)

    F = np.zeros(n, dtype=np.complex)

    F = np.r_[AB[0], (AB[1:nfft + 1] + 1j * AB[nfft + 1:]), np.zeros(nfft) ]
    return F

def rx2F(y, max_nfft, x=None):
    """Convert non-complex signal, y, to single sided Fourier components, that satifies x(t) = sum(X(cos(iw)+sin(iw)), i=0..N)"""
    d = np.arange(360)
    if x is not None:
        x,fit = bin_fit(x,y, d)
        y = fit(d)
    F = np.fft.rfft(y) / len(y)
    F[1:-1] *= 2  # add negative side
    F = np.conj(F)
    return F[:max_nfft + 1]

def rF2x(rF):
    """Convert single sided Fourier components, that satisfies x(t) = sum(X(cos(iw)+sin(iw)), i=0..N) to non-complex signal"""
    rF = np.conj(rF)
    rF[1:] /= 2
    rF = np.r_[rF, np.zeros(181 - len(rF), dtype=np.complex)]
    return np.fft.irfft(rF) * 360



'''
Created on 01/08/2016

@author: MMPE
'''
from wetb.hawc2.pc_file import PCFile
from wetb.hawc2.ae_file import AEFile
from wetb.hawc2.htc_file import HTCFile
import os
import numpy as np
from wetb.hawc2.st_file import StFile


class MainBody():
    def __init__(self, htcfile, body_name):
        if isinstance(htcfile, str):
            htcfile = HTCFile(htcfile)
        self.htcfile = htcfile
        s = htcfile.new_htc_structure
        main_bodies = {s[k].name[0]:s[k] for k in s.keys() if s[k].name_ == "main_body"}
        self.main_body = main_bodies[body_name]
        self.main_body = s.get_subsection_by_name(body_name)
        self.stFile = StFile(os.path.join(htcfile.modelpath, self.main_body.timoschenko_input.filename[0]))
        #self.c2def = np.array([v.values[1:5] for v in self.main_body.c2_def if v.name_ == "sec"])
        self.concentrated_mass = [cm.values for cm in self.main_body if cm.name_.startswith('concentrated_mass')]
        
    @property
    def c2def(self):
        if not hasattr(self, "_c2def"):
            self._c2def = np.array([v.values[1:5] for v in self.main_body.c2_def if v.name_ == "sec"])
        return self._c2def

    def plot_xz_geometry(self, plt=None):
        if plt is None:
            import matplotlib.pyplot as plt
            plt.figure()
        plt.xlabel("z")
        plt.ylabel("x")
        z = np.linspace(self.c2def[0, 2], self.c2def[-1, 2], 100)
        plt.plot(self.c2def[:, 2], self.c2def[:, 0],'.-', label='Center line')
        plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 0]) + self.stFile.x_e(z), label='Elastic center')
        plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 0]) + self.stFile.x_cg(z), label='Mass center')
        plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 0]) + self.stFile.x_sh(z), label='Shear center')
        for cm in self.concentrated_mass:
            plt.plot(self.c2def[cm[0]-1,2]+cm[3],self.c2def[cm[0]-1,0]+cm[1],'x', label='Concentrated mass')
        plt.legend()
    
    def plot_yz_geometry(self, plt=None):
        if plt is None:
            import matplotlib.pyplot as plt
            plt.figure()
        plt.xlabel("z")
        plt.ylabel("y")
        z = np.linspace(self.c2def[0, 2], self.c2def[-1, 2], 100)
        plt.plot(self.c2def[:, 2], self.c2def[:, 1], '.-', label='Center line')
        plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 1]) + self.stFile.y_e(z), label='Elastic center')
        plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 1]) + self.stFile.y_cg(z), label='Mass center')
        plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 1]) + self.stFile.y_sh(z), label='Shear center')
        for cm in self.concentrated_mass:
            plt.plot(self.c2def[cm[0]-1,2]+cm[3],self.c2def[cm[0]-1,1]+cm[2],'x', label='Concentrated mass')
        plt.legend()

        
# 
# 
# class BladeData(object):
#     def plot_xz_geometry(self, plt):
# 
#         z = np.linspace(self.c2def[0, 2], self.c2def[-1, 2], 100)
#         plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 0]), label='Center line')
#         plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 0]) + self.pcFile.chord(z) / 2, label='Leading edge')
#         plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 0]) - self.pcFile.chord(z) / 2, label="Trailing edge")
#         x, y, z = self.hawc2_splines()
#         #plt.plot(z, x, label='Hawc2spline')
# 
#     def plot_yz_geometry(self, plt):
# 
#         z = np.linspace(self.c2def[0, 2], self.c2def[-1, 2], 100)
#         plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 1]), label='Center line')
#         plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 1]) + self.pcFile.thickness(z) / 100 * self.pcFile.chord(z) / 2, label='Suction side')
#         plt.plot(z, np.interp(z, self.c2def[:, 2], self.c2def[:, 1]) - self.pcFile.thickness(z) / 100 * self.pcFile.chord(z) / 2, label="Pressure side")
#         x, y, z = self.hawc2_splines()
#         #plt.plot(z, y, label='Hawc2spline')
# 
#     def hawc2_splines(self):
#         curve_z = np.r_[0, np.cumsum(np.sqrt(np.sum(np.diff(self.c2def[:, :3], 1, 0) ** 2, 1)))]
#         curve_z_nd = curve_z / curve_z[-1]
# 
#         def akima(x, y):
#             n = len(x)
#             var = np.zeros((n + 3))
#             z = np.zeros((n))
#             co = np.zeros((n, 4))
#             for i in range(n - 1):
#                 var[i + 2] = (y[i + 1] - y[i]) / (x[i + 1] - x[i])
#             var[n + 1] = 2 * var[n] - var[n - 1]
#             var[n + 2] = 2 * var[n + 1] - var[n]
#             var[1] = 2 * var[2] - var[3]
#             var[0] = 2 * var[1] - var[2]
# 
#             for i in range(n):
#                 wi1 = abs(var[i + 3] - var[i + 2])
#                 wi = abs(var[i + 1] - var[i])
#                 if (wi1 + wi) == 0:
#                     z[i] = (var[i + 2] + var[i + 1]) / 2
#                 else:
#                     z[i] = (wi1 * var[i + 1] + wi * var[i + 2]) / (wi1 + wi)
# 
#             for i in range(n - 1):
#                 dx = x[i + 1] - x[i]
#                 a = (z[i + 1] - z[i]) * dx
#                 b = y[i + 1] - y[i] - z[i] * dx
#                 co[i, 0] = y[i]
#                 co[i, 1] = z[i]
#                 co[i, 2] = (3 * var[i + 2] - 2 * z[i] - z[i + 1]) / dx
#                 co[i, 3] = (z[i] + z[i + 1] - 2 * var[i + 2]) / dx ** 2
#             co[n - 1, 0] = y[n - 1]
#             co[n - 1, 1] = z[n - 1]
#             co[n - 1, 2] = 0
#             co[n - 1, 3] = 0
#             return co
# 
#         def coef2spline(s, co):
#             x, y = [], []
#             for i, c in enumerate(co.tolist()[:-1]):
#                 p = np.poly1d(c[::-1])
#                 z = np.linspace(0, s[i + 1] - s[i ], 10)
#                 x.extend(s[i] + z)
#                 y.extend(p(z))
#             return y
# 
#         x, y, z = [coef2spline(curve_z_nd, akima(curve_z_nd, self.c2def[:, i])) for i in range(3)]
#         return x, y, z



# class Blade(MainBody, BladeData):
#     def __init__(self, htc_filename, modelpath=None, blade_number=1):
#         
#         self.htcfile = htcfile = HTCFile(htc_filename, modelpath)
#         
#         blade_name = [link[2] for link in htcfile.aero if link.name_.startswith('link') and link[0]==blade_number][0]
#         MainBody.__init__(self, htc_filename, modelpath, blade_name)
#         self.pcFile = PCFile(os.path.join(htcfile.modelpath, htcfile.aero.pc_filename[0]),
#                         os.path.join(htcfile.modelpath, htcfile.aero.ae_filename[0]))
#         
#     def plot_xz_geometry(self, plt=None):
#         if plt is None:
#             import matplotlib.pyplot as plt
#             plt.figure()
# 
#         MainBody.plot_xz_geometry(self, plt)
#         BladeData.plot_xz_geometry(self, plt=plt)
#         plt.legend()
#     
#     def plot_geometry_yz(self, plt=None):
#         if plt is None:
#             import matplotlib.pyplot as plt
#             plt.figure()
# 
#         BladeData.plot_yz_geometry(self, plt=plt)
#         MainBody.plot_yz_geometry(self, plt)
#         






'''
Created on 01/10/2014

@author: MMPE
'''
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import int
from builtins import map
from builtins import str
from builtins import zip
from future import standard_library
standard_library.install_aliases()
import pandas as pd
import numpy as np
import glob
import os
import functools

from wetb.hawc2.sel_file import SelFile
from wetb.utils.caching import cache_function
from collections import OrderedDict
#HOURS_PR_20YEAR = 20 * 365 * 24

# hack around FileNotFoundError not being in Python2
try:
    FileNotFoundError
except NameError as e:
    class FileNotFoundError(OSError):
        pass


def Weibull(u, k, start, stop, step):
    C = 2 * u / np.sqrt(np.pi)
    cdf = lambda x :-np.exp(-(x / C) ** k)

    return {wsp:-cdf(wsp - step / 2) + cdf(wsp + step / 2) for wsp in np.arange(start, stop + step, step)}

def Weibull2(u, k, wsp_lst):
    C = 2 * u / np.sqrt(np.pi)
    cdf = lambda x :-np.exp(-(x / C) ** k)
    edges = np.r_[wsp_lst[0] - (wsp_lst[1] - wsp_lst[0]) / 2, (wsp_lst[1:] + wsp_lst[:-1]) / 2, wsp_lst[-1] + (wsp_lst[-1] - wsp_lst[-2]) / 2]
    return [-cdf(e1) + cdf(e2) for wsp, e1, e2 in zip(wsp_lst, edges[:-1], edges[1:])]

def Weibull_IEC(Vref, Vhub_lst):
    """Weibull distribution according to IEC 61400-1:2005, page 24

    Parameters
    ----------
    Vref : int or float
        Vref of wind turbine class
    Vhub_lst : array_like
        Wind speed at hub height. Must be equally spaced.

    Returns
    -------
    nd_array : list of probabilities

    Examples
    --------
    >>> Weibull_IEC(50, [4,6,8])
    [ 0.11002961  0.14116891  0.15124155]
    """
    Vhub_lst = np.array(Vhub_lst)
    #Average wind speed
    Vave=.2*Vref
    #Rayleigh distribution
    Pr = lambda x : 1 - np.exp(-np.pi*(x/(2*Vave))**2)
    #Wsp bin edges: [4,6,8] -> [3,5,7,9]
    wsp_bin_edges = np.r_[Vhub_lst[0] - (Vhub_lst[1] - Vhub_lst[0]) / 2, (Vhub_lst[1:] + Vhub_lst[:-1]) / 2, Vhub_lst[-1] + (Vhub_lst[-1] - Vhub_lst[-2]) / 2]
    #probabilities of 3-5, 5-7, 7-9
    return np.array([-Pr(e1) + Pr(e2) for e1, e2 in zip(wsp_bin_edges[:-1], wsp_bin_edges[1:])])



class DLCHighLevel(object):

    def __init__(self, filename, fail_on_resfile_not_found=False, shape_k=2.0):
        self.filename = filename
        self.fail_on_resfile_not_found = fail_on_resfile_not_found

        # Weibul distribution shape parameter
        self.shape_k = shape_k

        # Variables
        df_vars = pd.read_excel(self.filename, sheetname='Variables',
                                index_col='Name')
        df_vars.fillna('', inplace=True)
        for name, value in zip(df_vars.index, df_vars.Value.values):
            setattr(self, name.lower(), value)
        if not hasattr(self, "res_path"):
            raise Warning("The 'Variables' sheet of '%s' must contain the "
                          "variable 'res_path' specifying the path to the "
                          "result folder" % self.filename)
        self.res_path = os.path.join(os.path.dirname(self.filename), self.res_path)

        #DLC sheet
        self.dlc_df = pd.read_excel(self.filename, sheetname='DLC', skiprows=[1])
        # empty strings are now nans, convert back to empty strings
        self.dlc_df.fillna('', inplace=True)
        # force headers to lower case
        self.dlc_df.columns = [k.lower() for k in self.dlc_df.columns]
        if 'dlc' not in self.dlc_df.columns and 'name' in self.dlc_df.columns:
            # rename old style "name" column to "dlc"
            self.dlc_df = self.dlc_df.rename(columns={'name': 'dlc'})
        # ignore rows where column dlc is empty
        self.dlc_df = self.dlc_df[self.dlc_df['dlc'] != '']
        for k in ['load', 'dlc', 'dlc_dist', 'wsp', 'wsp_dist']:
            assert k.lower() in self.dlc_df.keys(), "DLC sheet must have a '%s' column" % k
        self.dist_value_keys = [('dlc_dist', 'dlc'), ('wsp_dist', 'wsp')]
        self.dist_value_keys.extend([(k, k.replace("_dist", ""))
                                     for k in self.dlc_df.keys()
                                     if k.endswith("_dist") and k not in ('dlc_dist', 'wsp_dist')])
        for i, (dk, vk) in enumerate(self.dist_value_keys):
            try:
                msg = "DLC sheet must have a '%s'-column when having a '%s'-column"
                assert vk in self.dlc_df.keys(), msg % (vk, dk)
            except AssertionError as e:
                if vk == "dlc" and 'name' in self.dlc_df.keys():
                    columns = list(self.dlc_df.columns)
                    columns[columns.index('name')] = 'dlc'
                    self.dlc_df.columns = columns
                else:
                    raise e
            self.dlc_df[vk].values[:] = [str(n).lower().replace(vk, "") for n in self.dlc_df[vk]]
        if 'psf' not in self.dlc_df:
            self.dlc_df['psf'] = 1

        # Sensors sheet
        self.sensor_df = pd.read_excel(self.filename, sheetname='Sensors')
        # empty strings are now nans, convert back to empty strings
        self.sensor_df.fillna('', inplace=True)
        # force headers to lower case
        self.sensor_df.columns = [k.lower() for k in self.sensor_df.columns]

        for k in ['Name', 'Nr']:
            assert k.lower() in self.sensor_df.keys(), "Sensor sheet must have a '%s' column" % k
        self.sensor_df = self.sensor_df[self.sensor_df.name!=""]
        assert not any(self.sensor_df['name'].duplicated()), "Duplicate sensor names: %s" % ",".join(self.sensor_df['name'][self.sensor_df['name'].duplicated()].values)
        for k in ['description', 'unit', 'statistic', 'ultimate', 'fatigue', 'm', 'neql', 'extremeload', 'bearingdamage', 'mindistance', 'maxdistance']:
            if k not in self.sensor_df.keys():
                self.sensor_df[k] = ""
        for _, row in self.sensor_df[self.sensor_df['fatigue'] != ""].iterrows():
            msg = "Invalid m-value for %s (m='%s')" % (row['name'], row['m'])
            assert isinstance(row['m'], (int, float)), msg
            msg = "Invalid NeqL-value for %s (NeqL='%s')" % (row['name'], row['neql'])
            assert isinstance(row['neql'], (int, float)), msg
        for name, nrs in zip(self.sensor_info("extremeload").name, self.sensor_info("extremeload").nr):
            msg = "'Nr' for Extremeload-sensor '%s' must contain 6 sensors (Fx,Fy,Fz,Mx,My,Mz)" % name
            assert (np.atleast_1d((eval(str(nrs)))).shape[0] == 6), msg

    def __str__(self):
        return self.filename

    def sensor_info(self, sensors=[]):
        if sensors != []:
            sensors = np.atleast_1d(sensors)
            empty_column = pd.DataFrame([""] * len(self.sensor_df.name))[0]
            return self.sensor_df[functools.reduce(np.logical_or, [((self.sensor_df.get(f, empty_column).values != "") | (self.sensor_df.name == f)) for f in sensors])]
        else:
            return self.sensor_df

    def dlc_variables(self, dlc):
        dlc_row = self.dlc_df[self.dlc_df['name'] == dlc]
        def get_lst(x):
            if isinstance(x, pd.Series):
                x = x.iloc[0]
            if ":" in str(x):
                start, step, stop = [float(eval(v, globals(), self.__dict__)) for v in x.lower().split(":")]
                return list(np.arange(start, stop + step, step))
            else:
                return [float(eval(v, globals(), self.__dict__)) for v in str(x).lower().replace("/", ",").split(",")]
        wsp = get_lst(dlc_row['wsp'])
        wdir = get_lst(dlc_row['wdir'])
        return wsp, wdir

    def distribution(self, value_key, dist_key, row):
        values = self.dlc_df[value_key][row]
        if ":" in values:
            start, step, stop = [float(eval(v, globals(), self.__dict__)) for v in values.lower().split(":")]
            values = np.arange(start, stop + step, step)
        else:
            try:
                values = [(eval(v, globals(), self.__dict__)) for v in str(values).lower().replace("/", ",").split(",")]
            except SyntaxError:
                values = [(eval(v.lstrip('0'), globals(), self.__dict__)) for v in str(values).lower().replace("/", ",").split(",")]

        dist = self.dlc_df[dist_key][row]
        if str(dist).lower() == "weibull" or str(dist).lower() == "rayleigh":
            dist = Weibull_IEC(self.vref, values)
        else:
            def fmt(v):
                if "#" in str(v):
                    return v
                else:
                    if v == "":
                        return 0
                    else:
                        return float(v) / 100
            dist = [fmt(v) for v in str(self.dlc_df[dist_key][row]).replace("/", ",").split(",")]
        assert len(values) == len(dist), "Number of %s-values (%d)!= number of %s-values(%d)" % (value_key, len(values), dist_key, len(dist))
        return OrderedDict(zip(map(self.format_tag_value, values), dist))

    def fatigue_distribution(self):
        fatigue_dist = {}
        for row, load in self.dlc_df['load'].iteritems():
            if "F" not in str(load).upper():
                continue
            dlc = self.dlc_df[self.dist_value_keys[0][1]][row]
            fatigue_dist[str(dlc)] = [self.distribution(value_key, dist_key, row) for dist_key, value_key in self.dist_value_keys]
        return fatigue_dist

    def files_dict(self, files=None):
        """
        Parameters
        ----------

        files : list, default=None
            When files is None, files_dict will search for files defined in
            the res_folder or res_path attribute if the former is absence.

        Returns
        -------

        files_dict : dict
            Dictionary holding the file name, total run hours as key, value
            pairs.
        """
        fatigue_dlcs = self.dlc_df[['F' in str(l).upper() for l in self.dlc_df['load']]]['dlc']
        if len(fatigue_dlcs) == 0:
            return {}
        ext = getattr(self, 'res_ext', ".sel")
        if isinstance(files, list):
            pass
        elif not hasattr(self, "res_folder") or self.res_folder == "":
            files = glob.glob(os.path.join(self.res_path, "*"+ext)) + glob.glob(os.path.join(self.res_path, "*/*"+ext))
            if len(files)==0:
                raise Exception('No *%s files found in:\n%s or\n%s'%(ext, self.res_path, os.path.join(self.res_path, "*/")))
        else:
            files = []
            for dlc_id in fatigue_dlcs:
                dlc_id = str(dlc_id)
                if "%" in self.res_folder:
                    folder = self.res_folder % dlc_id
                else:
                    folder = self.res_folder
                dlc_files = (glob.glob(os.path.join(self.res_path , folder, "*"+ext)))
                if len(dlc_files)==0:
                    raise Exception('DLC%s included in fatigue analysis, but no *%s files found in:\n%s'%(dlc_id, ext, os.path.join(self.res_path , folder)))
                files.extend(dlc_files)
        keys = list(zip(*self.dist_value_keys))[1]
        fmt = self.format_tag_value
        tags = [[fmt(tag.replace(key, "")) for tag, key in zip(os.path.basename(f).split("_"), keys)] for f in files]
        dlc_tags = list(zip(*tags))[0]
        files_dict = {dlc_tag:{} for dlc_tag in dlc_tags}
        for tag_row, f in zip(tags, files):
            d = files_dict[tag_row[0]]
            for tag in tag_row[1:]:
                if tag not in d:
                    d[tag] = {}
                d = d[tag]
            if 'files' not in d:
                d['files'] = []
            d['files'].append(f)
        return files_dict

    def format_tag_value(self, v):
        try:
            if int(float(v)) == float(v):
                return int(float(v))
            return float(v)
        except ValueError:
            return v

    def probability(self, props, f, files):
        total_prop = 1
        for prop in props[::-1]:
            if str(prop).startswith("#"):
                duration = SelFile(f).duration
                prop = float(prop[1:]) * duration / (60 * 60 * 24 * 365)
                return prop * total_prop
            else:
                total_prop *= prop
        return total_prop

    def file_hour_lst(self, years=20, files_dict=None, dist_dict=None, files=None):
        """Create a list of (filename, hours_pr_year) that can be used as input for LifeTimeEqLoad

        Returns
        -------
        file_hour_lst : list
            [(filename, hours),...] where\n
            - filename is the name of the file, including path
            - hours is the number of hours pr. 20 year of this file
        """

        fh_lst = []
        if dist_dict is None:
            dist_dict = self.fatigue_distribution()
        if files_dict is None:
            files_dict = self.files_dict(files=files)

        for dlc_id in sorted(dist_dict.keys()):
            dlc_id = str(dlc_id)

            fmt = self.format_tag_value
            def tag_prop_lst(dist_lst):
                if len(dist_lst) == 0:
                    return [[]]
                return [[(fmt(tag), prop)] + tl for tl in tag_prop_lst(dist_lst[1:]) for tag, prop in dist_lst[0].items()]

            def files_from_tags(self, f_dict, tags):
                if len(tags) == 0:
                    return f_dict['files']
                try:
                    return files_from_tags(self, f_dict[tags[0]], tags[1:])
                except KeyError:
                    if self.dist_value_keys[-len(tags)][1] == "wdir":
                        try:
                            return files_from_tags(self, f_dict[tags[0] % 360], tags[1:])
                        except:
                            pass
                    raise

            for tag_props in (tag_prop_lst(dist_dict[dlc_id])):
                tags, props = zip(*tag_props)
                try:
                    files = (files_from_tags(self, files_dict, tags))
                except KeyError:
                    if self.fail_on_resfile_not_found:
                        raise FileNotFoundError("Result files for %s not found" % (", ".join(["%s='%s'" % (dv[1], t) for dv, t in zip(self.dist_value_keys, tags)])))
                    else:
                        continue
                if files:
                    f_prob = self.probability(props, files[0], files) / len(files)
                    f_hours_pr_20year = 365 * 24 * years * f_prob
                    for f in sorted(files):
                        fh_lst.append((f, f_hours_pr_20year))
        return fh_lst

    def dlc_lst(self, load='all'):
        dlc_lst = np.array(self.dlc_df['dlc'])[np.array([load == 'all' or load.lower() in d.lower() for d in self.dlc_df['load']])]
        return [v.lower().replace('dlc', '') for v in dlc_lst]

    @cache_function
    def psf(self):
        return {dlc: float((psf, 1)[psf == ""]) for dlc, psf in zip(self.dlc_df['dlc'], self.dlc_df['psf']) if dlc != ""}


if __name__ == "__main__":
    dlc_hl = DLCHighLevel(r'X:\DTU10MW\Q0010\DLC_post_betas1.xlsx')
    #print (DLCHighLevelInputFile(r'C:\mmpe\Projects\DLC.xlsx').sensor_info(0, 0, 1)['Name'])
    #print (dlc_dict()['64'])
    #print (dlc_hl.fatigue_distribution()['64'])
    print (dlc_hl.file_hour_lst(r"X:\DTU10MW/Q0010/res/"))





import os
import sys

from unittest import mock
import pytest
import matplotlib.pyplot as plt


def run_module_main(module):
    # check that all main module examples run without errors
    if os.name == 'posix' and "DISPLAY" not in os.environ:
        pytest.xfail("No display")

    def no_show(*args, **kwargs):
        pass
    plt.show = no_show  # disable plt show that requires the user to close the plot

    def no_print(s):
        pass

    try:
        with mock.patch.object(module, "__name__", "__main__"):
            with mock.patch.object(module, "print", no_print):
                getattr(module, 'main')()
    except Exception as e:
        raise type(e)(str(e) +
                      ' in %s.main' % module.__name__).with_traceback(sys.exc_info()[2])



'''
Created on 11/07/2016

@author: MMPE
'''
import unittest
import os

tfp = os.path.join(os.path.dirname(__file__), 'test_files/')
class Test(unittest.TestCase):
    pass
        


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testUnix_filename']
    unittest.main()



'''
Created on 3. maj 2017

@author: mmpe
'''
import os
import unittest

from wetb.hawc2.blade import H2Blade
import numpy as np

tfp = os.path.join(os.path.dirname(__file__), 'test_files/')  # test file path
class Test(unittest.TestCase):


#     def testBladeInfo(self):
#         bi = H2Blade(tfp + "simulation_setup/DTU10MWRef6.0/htc/DTU_10MW_RWT.htc")
#         if 0:
#             import matplotlib.pyplot as plt
#             print (dir(bi))
#             #print (bi.radius_s())
#             plt.plot(bi.radius_s(), bi.twist())
#             plt.plot(bi.c2def[:,2], bi.c2def[:,3])
#             x = np.linspace(0,1,1000)
#             plt.plot(bi.blade_radius*x, bi.c2nd(x)[:,3])
#             plt.show()

            
    def testBladeInfo_AE(self):
        bi = H2Blade(None, tfp + "NREL_5MW_ae.txt", tfp + "NREL_5MW_pc.txt")
        self.assertEqual(bi.thickness(32), 23.78048780487805)
        self.assertEqual(bi.chord(32), 3.673)
        self.assertEqual(bi.pc_set_nr(32), 1)

    def testBladeInfo_PC_AE(self):
        bi = H2Blade(None, tfp + "NREL_5MW_ae.txt", tfp + "NREL_5MW_pc.txt")
        self.assertEqual(bi.CL(36, 10), 1.358)
        self.assertEqual(bi.CD(36, 10), 0.0255)
        self.assertEqual(bi.CM(36, 10), -0.1103)
            
#     def test_curved_length2radius(self):
#         bi = H2BladeInfo(tfp + "simulation_setup/DTU10MWRef6.0/htc/DTU_10MW_RWT.htc")
#         np.testing.assert_array_equal(bi.xyztwist(86.4979, curved_length=True), bi.xyztwist(86.3655))
        
            
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testBladeInfo']
    unittest.main()